import base64
from collections import namedtuple
import time

from traits.api import (Any, Bool, Dict, Enum, Event, HasTraits,
                        Instance, Int, List, Callable)
import vtk

from tvtk.api import tvtk
from ..figure import figure, gcf
from ..engine_manager import options
from tvtk.tvtk_base import global_disable_update

options.offscreen = True

EventInfo = namedtuple('EventInfo', ['id', 'name', 'event', 'data'])


if hasattr(vtk, 'vtkOSOpenGLRenderWindow'):
    # Needed for VTK's vtkglew to be able to load the OSMesa.so. See:
    # https://www.vtk.org/pipermail/vtk-developers/2017-November/035592.html
    import ctypes
    osm = ctypes.CDLL("libOSMesa.so", ctypes.RTLD_GLOBAL)


class ImageEncoder(HasTraits):
    scene = Any
    w2if = Instance(tvtk.WindowToImageFilter)
    writer = Instance(tvtk.ImageWriter)
    image_type = Enum('png', 'jpeg')
    quality = Int(60)
    compress = Bool(False)

    _png_writer = Instance(tvtk.ImageWriter)
    _jpg_writer = Instance(tvtk.ImageWriter)

    # ---- Public protocol -------
    def get_raw_image(self):
        '''Returns the raw bytes of the image.
        '''
        self.w2if.modified()
        w = self.writer
        w.update()
        w.write()
        return w.result.to_array().tostring()

    # ---- Private protocol -------
    def _w2if_default(self):
        w2if = tvtk.WindowToImageFilter()
        if self.scene is not None:
            w2if.input = self.scene.render_window
        return w2if

    def _writer_default(self):
        return self._get_writer(self.image_type)

    def _get_writer(self, image_type):
        if image_type == 'png':
            writer = self._png_writer
        else:
            writer = self._jpg_writer
        writer.set_input_connection(self.w2if.output_port)
        return writer

    def __png_writer_default(self):
        return tvtk.PNGWriter(write_to_memory=True)

    def __jpg_writer_default(self):
        return tvtk.JPEGWriter(quality=self.quality, write_to_memory=True)

    def _quality_changed(self, value):
        self._jpg_writer.quality = value

    def _image_type_changed(self, value):
        if self.writer is not None:
            self.writer.set_input_connection(None)
        self.writer = self._get_writer(value)

    def _scene_changed(self, value):
        self.w2if.input = value.render_window

    def _compress_changed(self, value):
        if value:
            self.image_type = 'jpeg'
        else:
            self.image_type = 'png'


class RemoteScene(HasTraits):

    #: Any events for clients to listen to. Only a subset of events are emitted
    # currently, the event is an `EventInfo` instance.
    event = Event

    #: Set a callback function with signature `f(secs, callable, *args, **kw)`.
    # The function should call the passed `callable` function with the given
    # args and kwargs after the specified time in seconds.
    call_later = Callable

    #: The image encoder which converts the scene to a suitable image.
    image_encoder = Instance(ImageEncoder)

    def __init__(self, figure=None, **traits):
        super(RemoteScene, self).__init__(**traits)
        if figure is None:
            figure = gcf()
        self.figure = figure
        self.scene = figure.scene
        self.trw = self.scene.render_window
        self.trwi = self.scene.interactor
        self.rwi = tvtk.to_vtk(self.trwi)
        self.rw = tvtk.to_vtk(self.trw)
        self.ren = tvtk.to_vtk(self.scene.renderer)
        self.id = id(self)
        self._last_image = ''
        self._last_render = 0
        self._time_to_render = 0
        self._time_for_image = 0
        self._render_size = 200*200
        self._pending_render = False
        self._doing_render = False
        self._timer_enabled = True
        self._setup_scene()

    # ## Public protocol ####
    def get_raw_image(self):
        # This flag is set to prevent render events from being generated
        # while generating the image (which does happen!).
        self._doing_render = True
        data = self.image_encoder.get_raw_image()
        self._doing_render = False
        return data

    def get_image(self):
        data = self.get_raw_image()
        return base64.encodebytes(data).decode('ascii')

    def call_rwi(self, method, *args):
        if method == 'SetSize':
            # This is an issue with the way TVTK is setup and how VTK resizes
            # an offscreen window. It basically releases all graphics resources
            # and this job is passed on to the various mappers. When the mapper
            # does this it emits a ModifiedEvent which is picked by TVTK to
            # fire a render. This call to render when the window is resizing
            # locks up VTK or entirely messes up the rendering. I've reported
            # this upstream but it turns out that disabling the auto updates in
            # TVTK maybe of general use so we have a convenient context manager
            # to do this called `global_disable_update` that shuts off the
            # automatic updates for ALL TVTK objects for a brief while. This
            # fixes the issue.
            with global_disable_update():
                self.trw.size = args
                self.trw.render()

        if 'PressEvent' in method:
            self.image_encoder.compress = True
        if 'ReleaseEvent' in method:
            self.image_encoder.compress = False

        getattr(self.rwi, method)(*args)

    # ## Private protocol ####

    def _send_pending_render(self):
        if self._pending_render:
            self._send_render_event()

    def _emit_event(self, obj, evt):
        self.event = EventInfo(self.id, obj.GetClassName(), evt, None)

    def _required_time_to_render(self):
        return (self._time_for_image + self._time_to_render +
                self._render_size/1e6)

    def _ok_to_send_render_event(self):
        required_time = self._required_time_to_render()
        return (time.time() - self._last_render) > required_time

    def _on_create_timer(self, obj, evt):
        self._timer_enabled = True
        self.call_later(0.05, self._on_timer_event)

    def _on_destroy_timer(self, obj, evt):
        self._timer_enabled = False

    def _on_timer_event(self):
        if self._timer_enabled:
            self.rwi.TimerEvent()

    @vtk.calldata_type(vtk.VTK_INT)
    def _on_cursor_changed_event(self, obj, evt, calldata):
        self.event = EventInfo(self.id, obj.GetClassName(), evt, calldata)

    def _on_render_event(self, obj, event):
        if self._doing_render:
            return

        if self._ok_to_send_render_event():
            self._send_render_event()
        else:
            if not self._pending_render and self.call_later:
                req_time = self._required_time_to_render()
                # print("Required time:", req_time)
                self.call_later(min(req_time, 0.25), self._send_pending_render)
            self._pending_render = True

    def _send_render_event(self):
        self._pending_render = False
        event = 'RenderEvent'
        start = time.time()
        img = self.get_image()
        self._render_size = len(img)
        # print(self._render_size)
        self._last_render = time.time()
        self._time_for_image = self._last_render - start
        self._time_to_render = self.ren.GetLastRenderTimeInSeconds()
        # print("Time to render:", self._time_to_render)
        if img != self._last_image:
            self._last_image = img
            format = self.image_encoder.image_type.upper()
            data = dict(data=img, type='image', format=format,
                        time=self._time_for_image,
                        render_time=self._time_to_render)
            self.event = EventInfo(
                self.id, self.rw.GetClassName(), event, data
            )

    def _setup_scene(self):
        self.trwi.interactor_style.set_current_style_to_trackball_camera()
        self.rwi.AddObserver('CreateTimerEvent', self._on_create_timer)
        self.rwi.AddObserver('DestroyTimerEvent', self._on_destroy_timer)
        rw = self.rw
        rw.AddObserver('RenderEvent', self._on_render_event)
        rw.AddObserver('CursorChangedEvent', self._on_cursor_changed_event)

    def _image_encoder_default(self):
        return ImageEncoder(scene=self.scene, quality=60)


class SceneManager(HasTraits):

    scenes = Dict

    figure_to_id = Dict

    clients = List

    #: Set a callback function with signature `f(secs, callable, *args, **kw)`.
    # The function should call the passed `callable` function with the given
    # args and kwargs after the specified time in seconds.
    call_later = Callable

    def add_client(self, client):
        '''All the client needs to do is to provide a method called
        `handle_event` which is given the event data from the scenes
        and route that to the appropriate recipient.

        The event data is basically an `EventInfo` instance.
        '''
        self.clients.append(client)

    def remove_client(self, client):
        self.clients.remove(client)

    def register_figure(self, figure):
        '''Given an existing figure, set it up for remote visualization.
        '''
        remote = RemoteScene(figure=figure)
        r_id = id(remote)
        self.figure_to_id[figure] = r_id
        self.scenes[r_id] = remote
        self._setup_events(remote)
        if self.call_later:
            remote.call_later = self.call_later
        return r_id

    def figure(self, *args, **kw):
        '''Create a figure and manage it as a remote figure. Any args
        and kwargs are passed on to the `mlab.figure` function.

        '''
        f = figure(*args, **kw)
        return self.register_figure(f)

    def message(self, obj_id, method_name, *args, **kw):
        '''Call a method with given args but do not return any responses.

        This is meant to be used for passing messages to the scenes.
        '''
        self.call(obj_id, method_name, *args, **kw)

    def call(self, obj_id, method_name, *args, **kw):
        '''Call a method and also return the response of the method.

        Used for synchronous calls to the scenes.
        '''
        scene = self.scenes[obj_id]
        method = getattr(scene, method_name)
        return method(*args, **kw)

    # ## Private protocol ####

    def _setup_events(self, obj):
        obj.on_trait_change(self._forward_event, 'event')

    def _forward_event(self, event):
        # Override this to change the default if needed.
        for client in self.clients:
            client.handle_event(event)

    def _call_later_changed(self, f):
        for scene in self.scenes.values():
            scene.call_later = f



"""The base object from which all MayaVi pipeline objects derive.

"""
# Author: Prabhu Ramachandran <prabhu_r@users.sf.net>
# Copyright (c) 2005-2020, Enthought, Inc.
# License: BSD Style.

# Standard library imports.
import pickle
from copy import deepcopy
import os
import logging
import importlib

# Enthought library imports.
from traits.api import (Any, Instance, Property, Bool, Str, Python,
                        HasTraits, WeakRef, on_trait_change)
from traitsui.api import TreeNodeObject
from tvtk.pyface.tvtk_scene import TVTKScene
from apptools.persistence import state_pickler
from pyface.resource.api import resource_path
from pyface.image_resource import ImageResource
from traitsui.menu import Menu, Action, Separator
from traitsui.api import View
from apptools.scripting.api import Recorder

# Local imports.
from mayavi.preferences.api import preference_manager
from mayavi.core.common import get_engine

# Setup a logger for this module.
logger = logging.getLogger(__name__)

# Subdirectory that the Base class will check for possible external views.
UI_DIR_NAME = ['ui']

# ------------------------------------------------------------------------------
#  The core tree node menu actions:
# ------------------------------------------------------------------------------

NewAction = 'NewAction'
CopyAction = Action(
    name='Copy',
    action='editor._menu_copy_node',
    enabled_when='editor._is_copyable(object)'
)
CutAction = Action(
    name='Cut',
    action='editor._menu_cut_node',
    enabled_when='editor._is_cutable(object)'
)
PasteAction = Action(
    name='Paste',
    action='editor._menu_paste_node',
    enabled_when='editor._is_pasteable(object)'
)
DeleteAction = Action(
    name='Delete',
    action='editor._menu_delete_node',
    enabled_when='editor._is_deletable(object)'
)
RenameAction = Action(
    name='Rename',
    action='editor._menu_rename_node',
    enabled_when='editor._is_renameable(object)'
)
standard_menu_actions = [
    Separator(),
    CutAction,
    CopyAction,
    PasteAction,
    Separator(),
    RenameAction,
    DeleteAction,
    Separator(),
]


######################################################################
# `Base` class.
######################################################################
class Base(TreeNodeObject):
    # The version of this class.  Used for persistence.
    __version__ = 0

    ########################################
    # Traits

    # The scene (RenderWindow) associated with this component.
    scene = Instance(TVTKScene, record=False)

    # Is this object running as part of the mayavi pipeline.
    running = Property(Bool, record=False)

    # The object's name.
    name = Str('')

    # The default icon.
    icon = 'module.ico'

    # The human readable type for this object
    type = Str('', record=False)

    # Is this object visible or not.
    visible = Bool(True, desc='if the object is visible')

    # Extend the children list with an AdderNode when a TreeEditor needs it.
    children_ui_list = Property(depends_on=['children'], record=False)

    # The parent of this object, i.e. self is an element of the parents
    # children.  If there is no notion of a parent/child relationship
    # this trait is None.
    parent = WeakRef(record=False)

    # A helper for the right click menus, context sensitivity etc.
    menu_helper = Instance(HasTraits, record=False)

    # Our recorder.
    recorder = Instance(Recorder, record=False)

    ##################################################
    # Private traits
    _is_running = Bool(False)

    # This is used to save the state of the object when it is not
    # running.  When the object "starts", the state is loaded.  This
    # is done because a stopped object will not have a meaningful VTK
    # pipeline setup, so setting its state will lead to all kinds of
    # errors.
    _saved_state = Any('')

    # Hide and show actions
    _HideShowAction = Instance(Action,
                               kw={'name': 'Hide/Show',
                                   'action': 'object._hideshow'}, )

    # The menu shown on right-click for this.
    _menu = Instance(Menu, transient=True)

    # Path to the icon for this object.
    _icon_path = Str()

    # Adder node: a dialog to add children to this object
    _adder_node_class = None

    # Name of the file that may host the hand-crafted view
    _view_filename = Str(transient=True)

    # Hand crafted view.
    _module_view = Instance(View, transient=True)

    # Work around problem with HasPrivateTraits.
    __ = Python
    ##################################################

    ######################################################################
    # `object` interface
    ######################################################################
    def __get_pure_state__(self):
        """Method used by the state_pickler.
        """
        d = self.__dict__.copy()
        for attr in ('scene', '_is_running', '__sync_trait__',
                     '__traits_listener__', '_icon_path',
                     '_menu', '_HideShowAction', 'menu_helper',
                     'parent', 'parent_', '_module_view', '_listener_cache',
                     '_view_filename', 'mlab_source'):
            d.pop(attr, None)
        return d

    def __getstate__(self):
        """Allows standard pickle to work via the state_pickler.
        """
        return state_pickler.dumps(self)

    def __setstate__(self, str_state):
        """Allows standard pickle to work via the state_pickler.
        """
        self.__init__()
        # Get the state from the string and update it.
        state = state_pickler.loads_state(str_state)
        state_pickler.update_state(state)
        # Save the state and load it if we are running.
        self._saved_state = pickle.dumps(state)
        if self.running:
            self._load_saved_state()

    def __deepcopy__(self, memo):
        """Method used by copy.deepcopy().  This also uses the
        state_pickler to work correctly.
        """
        # Create a new instance.
        new = self.__class__()
        # If we have a saved state, use it for the new instance.  If
        # not, get our state and save that.
        saved_state = self._saved_state
        if len(saved_state) == 0:
            state = state_pickler.get_state(self)
            # FIXME: This is for streamline seed point widget position which
            # does not get serialized correctly
            try:
                st = state.children[0].children[4]
                l_pos = st.seed.widget.position
                st.seed.widget.position = [pos.item() for pos in l_pos]
            except (IndexError, AttributeError):
                pass
            saved_state = pickle.dumps(state)
        new._saved_state = saved_state
        # In the unlikely case that a new instance is running, load
        # the saved state.
        if new.running:
            new._load_saved_state()
        return new

    ######################################################################
    # `Base` interface
    ######################################################################
    def start(self):
        """Invoked when this object is added to the mayavi pipeline.
        """
        self.running = True
        self._load_saved_state()

    def stop(self):
        """Invoked when this object is removed from the mayavi
        pipeline.
        """
        self.running = False

    def add_child(self, child):
        """This method intelligently adds a child to this object in
        the MayaVi pipeline.
        """
        raise NotImplementedError

    def remove_child(self, child):
        """Remove specified child from our children.
        """
        raise NotImplementedError()

    def remove(self):
        """Remove ourselves from the mayavi pipeline.
        """
        if self.parent is not None:
            e = get_engine(self)
            self.parent.remove_child(self)
            if e.current_object is self:
                e.current_object = self.parent

    def render(self):
        """Invokes render on the scene, this in turn invokes Render on
        the VTK pipeline.
        """
        s = self.scene
        if s is not None:
            s.render()

    def dialog_view(self):
        """ Returns a view with an icon and a title.
        """
        view = self.trait_view()
        icon = self._icon_path + os.sep + 'images' + os.sep + self.icon
        view.icon = ImageResource(icon)
        view.title = "Edit%s: %s" % (self.type, self.name)
        view.buttons = ['OK', 'Cancel']
        return view

    def trait_view(self, name=None, view_element=None):
        """ Gets or sets a ViewElement associated with an object's class.

        Overridden here to search for a separate file in the same directory
        for the view to use for this object. The view should be declared in
        the file named <class name>_view. If a file with this name is not
        found, the trait_view method on the base class will be called.
        """

        # If a name is specified, then call the HasTraits trait_view method
        # which will return (or assign) the *view_element* associated with
        # *name*.
        if name:
            return super(Base, self).trait_view(name, view_element)

        view = self._load_view_cached(name, view_element)
        # Uncomment this when developping views.
        # view = self._load_view_non_cached(name, view_element)
        return view

    ######################################################################
    # `TreeNodeObject` interface
    ######################################################################
    def tno_get_label(self, node):
        """Gets the label to display for a specified object.
        """
        if self.name == '':
            self.name = self.__class__.__name__
        return self.name

    def tno_get_view(self, node):
        """Gets the View to use when editing an object.
        """
        view = self.trait_view()
        view.kind = 'subpanel'
        return view

    def tno_confirm_delete(self, node):
        """Confirms that a specified object can be deleted or not.
        """
        if preference_manager.root.confirm_delete:
            return None
        else:
            return True

    def tno_get_menu(self, node):
        """ Returns the contextual pop-up menu.
        """
        if self._menu is None:
            return super(Base, self).tno_get_menu(node)
        return self._menu

    def tno_get_icon(self, node, is_expanded):
        return self.icon

    def tno_get_icon_path(self, node):
        return self._icon_path

    def tno_delete_child(self, node, index):
        if len(self.children_ui_list) > len(self.children):
            del self.children[index - 1]
        else:
            del self.children[index]

    def tno_append_child(self, node, child):
        """ Appends a child to the object's children.
        """
        self.children.append(child)

    def tno_insert_child(self, node, index, child):
        """ Inserts a child into the object's children.
        """
        if len(self.children_ui_list) > len(self.children):
            idx = index - 1
        else:
            idx = index
        self.children[idx:idx] = [child]

    ######################################################################
    # Non-public interface
    ######################################################################
    def _get_running(self):
        return self._is_running

    def _set_running(self, new):
        if self._is_running == new:
            return
        else:
            old = self._is_running
            self._is_running = new
            self.trait_property_changed('running', old, new)

    def _get_children_ui_list(self):
        """ Getter for Traits Property children_ui_list.

        For the base class, do not add anything to the children list.
        """
        if ((not preference_manager.root.show_helper_nodes or
             len(self.children) > 0)
            or self._adder_node_class is None
            or (not self.type == ' scene' and
                'none' in self.output_info.datasets)):
            # We can't use isinstance, as we would have circular
            # imports
            return self.children
        else:
            return [self._adder_node_class(object=self)]

    @on_trait_change('children[]')
    def _trigger_children_ui_list(self, old, new):
        """ Trigger a children_ui_list change when scenes changed.
        """
        self.trait_property_changed('children_ui_list', old, new)

    def _visible_changed(self, value):
        # A hack to set the name when the tree view is not active.
        # `self.name` is set only when tno_get_label is called and this
        # is never called when the tree view is not shown leading to an
        # empty name.
        if len(self.name) == 0:
            self.tno_get_label(None)
        if value:
            self.name = self.name.replace(' [Hidden]', '')
        else:
            n = self.name
            if ' [Hidden]' not in n:
                self.name = "%s [Hidden]" % n

    def _load_view_cached(self, name, view_element):
        """ Use a cached view for the object, for faster refresh.
        """
        if self._module_view is not None:
            view = self._module_view
        else:
            logger.debug("No view found for [%s] in [%s]. "
                         "Using the base class trait_view instead.",
                         self, self._view_filename)
            view = super(Base, self).trait_view(name, view_element)
        return view

    def _load_view_non_cached(self, name, view_element):
        """ Loads the view by execing a file. Useful when tweaking
            views.
        """
        result = {}
        view_filename = self._view_filename
        try:
            exec(compile(
                open(view_filename).read(), view_filename, 'exec'), {}, result
            )
            view = result['view']
        except IOError:
            logger.debug("No view found for [%s] in [%s]. "
                         "Using the base class trait_view instead.",
                         self, view_filename)
            view = super(Base, self).trait_view(name, view_element)
        return view

    def _hideshow(self):
        if self.visible:
            self.visible = False
        else:
            self.visible = True

    def _load_saved_state(self):
        """Load the saved state (if any) of this object.
        """
        saved_state = self._saved_state
        if len(saved_state) > 0:
            state = pickle.loads(saved_state)
            if hasattr(self, '__set_pure_state__'):
                self.__set_pure_state__(state)
            else:
                state_pickler.set_state(self, state)
            self._saved_state = ''

    def __view_filename_default(self):
        """ The name of the file that will host the view.
        """
        module = self.__module__.split('.')
        class_filename = module[-1] + '.py'
        module_dir_name = module[1:-1]
        base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        view_filename = os.path.join(*([base_dir] + module_dir_name
                                       + UI_DIR_NAME + [class_filename]))
        return view_filename

    def __module_view_default(self):
        """ Try to load a view for this object.
        """
        view_filename = self._view_filename
        if os.path.exists(view_filename):
            spec = importlib.util.spec_from_file_location(
                'view', view_filename
            )
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            view = module.view
        else:
            view = None
        return view

    def __menu_default(self):
        extras = []
        if self.menu_helper is not None:
            extras = self.menu_helper.actions + self._extra_menu_items()
        menu_actions = [Separator()] + extras + \
            [Separator(), self._HideShowAction, Separator()] + \
            deepcopy(standard_menu_actions)
        return Menu(*menu_actions)

    def __icon_path_default(self):
        return resource_path()

    def _extra_menu_items(self):
        """Override this to generate any new menu actions you want on
        the right click menu."""
        return []



"""
Custom nodes for a Tree Editor that provide views for adding various nodes
to the tree.
"""
# Authors: Judah De Paula <judah@enthought.com>
#          Prabhu Ramachandran <prabhu_r@users.sf.net>
# Copyright (c) 2008, Enthought, Inc.
# License: BSD Style.

# Enthought library imports.
from traits.api import (HasTraits, Str, Property, Any, Button,
                                  List, Instance, provides,
                                  ToolbarButton)
from traitsui.api import View, Item, Group,\
        TextEditor, TreeEditor, TreeNode, ListEditor, ITreeNode
from pyface.api import ImageResource
from pyface.resource.api import resource_path

# Local imports.
from .registry import registry

###############################################################################
# AdderNode class
###############################################################################
@provides(ITreeNode)
class AdderNode(TreeNode):
    """ Base class that will display a TreeNode to add items to the tree.
    """

    # String to be shown in the TreeEditor.
    label = Str('Base AdderNode')

    # Default tooltip for this class.
    tooltip = Str('Add an item')

    # The parent object that should be manipulated for adding children.
    object =  Any

    # Duck-typing is necessary since Mayavi assumes nodes always have scenes.
    scene = Property

    # Trait view to show in the Mayavi current object panel.
    view = View(Group(label='AdderNode'))

    def dialog_view(self):
        """ View shown by double-clicking on the node.  Same as in Base().
        """
        view = self.trait_view()
        view.buttons = [ ]
        view.title = self.label
        view.icon = ImageResource('add.ico')
        view.resizable = True
        view.width = 350
        view.height = 650
        return view

    def _get_scene(self):
        """ Trait Property getter for 'scene'.
        """
        object = self.object
        if isinstance(object, AdderNode):
            return None
        if object is not None:
            return object.scene
        else:
            return None

    #------------------------------------------------------------------------
    # The ITreeNode interface needed by the Qt tree_editor
    #------------------------------------------------------------------------

    def get_label(self):
        return self.label

    def get_icon(self, obj, is_expanded=False):
        return self.icon_name

    def get_icon_path(self):
        return resource_path()

    def get_tooltip(self):
        return self.tooltip

    def allows_children(self):
        return False

    def get_children_id(self, node=None):
        return []

    def when_label_changed(self, label_updated, remove):
        return

    def when_column_labels_change(self, listener, remove):
        return

###############################################################################
# SceneAdderNode class
###############################################################################
class SceneAdderNode(AdderNode):
    """ Subclass for adding Scene nodes to a Mayavi Engine node.
    """

    # String to be shown in the TreeEditor.
    label = Str('Add a new scene')

    # The name of the icon
    icon_name = Str('add_scene.png')

    # Button for the View.
    add_scene = Button('Add a new scene',
                      image=ImageResource('add_scene.png'))

    # Trait view to show in the Mayavi current object panel.
    view = View(Group(Item('add_scene', show_label=False, style='custom'),
                      label='Add a scene'))


    def _add_scene_fired(self):
        """ Trait handler for when the add_scene button is clicked.
        """
        self.object.new_scene()


###############################################################################
# DocumentedItem class
###############################################################################
class DocumentedItem(HasTraits):
    """ Container to hold a name and a documentation for an action.
    """

    # Name of the action
    name = Str

    # Button to trigger the action
    add = ToolbarButton('Add', orientation='horizontal',
                    image=ImageResource('add.ico'))

    # Object the action will apply on
    object = Any

    # Two lines documentation for the action
    documentation = Str

    view = View('_',
                Item('add', style='custom', show_label=False),
                Item('documentation', style='readonly',
                    editor=TextEditor(multi_line=True),
                    resizable=True,
                    show_label=False),
                )

    def _add_fired(self):
        """ Trait handler for when the add_source button is clicked in
            one of the sub objects in the list.
        """
        action = getattr(self.object.menu_helper, self.id)
        action()


def documented_item_factory(name='', documentation='',
                id='', object=None):
    """ Factory for creating a DocumentedItem with the right button
        label.
    """
    documentation = documentation.replace('\n', '')
    documentation = documentation.replace('  ', '')

    class MyDocumentedItem(DocumentedItem):
        add = ToolbarButton('%s' % name, orientation='horizontal',
                        image=ImageResource('add.ico'))

    return MyDocumentedItem(
                        name=name,
                        documentation=documentation,
                        id=id,
                        object=object)


###############################################################################
# ListAdderNode class
###############################################################################
class ListAdderNode(AdderNode):
    """ A node for adding object, with a list of objects to add generated
        from the registry.
    """

    # The list of items to display to the user.
    items_list = List(DocumentedItem)

    # A reference to the registry, to generate this list.
    items_list_source = List()

    # Selected item
    selected_item = Instance(DocumentedItem)

    # A reference to self, to allow to build the tree view.
    self = Instance(AdderNode)

    # The icon of the displayed objects
    icon_name = Str('add.ico')

    def _self_default(self):
        return self

    def default_traits_view(self):
        nodes = [TreeNode(node_for=[AdderNode],
                          label='name',
                          copy=False,
                          delete=False,
                          rename=False,
                          children='items_list',
                          ),
                 TreeNode(node_for=[DocumentedItem],
                          label='name',
                          copy=False,
                          delete=False,
                          rename=False,
                          icon_item=self.icon_name,
                          ),
                 ]

        tree_editor = TreeEditor(editable=False,
                                 hide_root=True,
                                 orientation='vertical',
                                 selected='object.selected_item',
                                 nodes=nodes,
                                 on_dclick='object._on_tree_dclick',
                                 )

        view = View(Item('self',
                            show_label=False,
                            editor=tree_editor,
                            resizable=True,
                            springy=True,
                            height=0.5),
                    Item('selected_item', style='custom', show_label=False,
                            height=0.5),
                    resizable=True)
        return view


    def _object_changed(self, value):
        """ Trait handler for when the self.object trait changes.
        """
        result = []
        if value is not None:
            # Don't need 'x', but do need to generate the actions.
            x = value.menu_helper.actions
            for src in self.items_list_source:
                if not self._is_action_suitable(value, src):
                    continue
                name = src.menu_name.replace('&','')
                result.append(
                        documented_item_factory(
                                name=name,
                                documentation=src.help,
                                id=src.id,
                                object=value)
                        )
        self.items_list = result


    def _is_action_suitable(self, object, src):
        """ Check that the action described by src can be applied on the
            given object.
        """
        if  hasattr(object.menu_helper, 'check_%s' % src.id) \
                and getattr(object.menu_helper, 'check_%s' % src.id)():
            return True
        else:
            return False

    def _on_tree_dclick(self, object):
        """ Called when an user double clicks on an item in the tree
            view.
        """
        object._add_fired()


###############################################################################
# SourceAdderNode class
###############################################################################
class SourceAdderNode(ListAdderNode):
    """ Tree node that presents a view to the user to add a scene source.
    """

    # Button for adding a data file, with automatic format checking.
    open_file = ToolbarButton('Load data from file',
                                orientation='horizontal',
                                image=ImageResource('file.png'))

    # A reference to the registry, to generate this list.
    items_list_source = [source for source in registry.sources
                         if len(source.extensions) == 0]

    # The string to display on the icon in the TreeEditor.
    label = 'Add Data Source'

    # The icon of the displayed objects
    icon_name = Str('source.ico')

    # Trait view to show in the Mayavi current object panel.
    def default_traits_view(self):
        return View(Group(Group(Item('open_file', style='custom'),
                      show_labels=False, show_border=False),
                      Item('items_list', style='readonly',
                            editor=ListEditor(style='custom')),
                      show_labels=False,
                      label='Add a data source'))

    def _open_file_fired(self):
        """ Trait handler for when the open_file button is clicked.
        """
        self.object.menu_helper.open_file_action()

    def _is_action_suitable(self, object, src):
        return True


###############################################################################
# ModuleAdderNode class
###############################################################################
class ModuleAdderNode(ListAdderNode):
    """ Tree node that presents a view to the user to add modules.
    """
    # String to be shown in the TreeEditor.
    label = Str('Add a visualization module')

    # The icon of the displayed objects
    icon_name = Str('module.ico')

    # A reference to the registry, to generate this list.
    items_list_source = registry.modules

    def _object_changed(self, value):
        if value is not None:
            value.menu_helper._build_filter_actions()
        ListAdderNode._object_changed(self, value)


###############################################################################
# FilterAdderNode class
###############################################################################
class FilterAdderNode(ListAdderNode):
    """ Tree node that presents a view to the user to add filters.
    """
    # String to be shown in the TreeEditor.
    label = Str('Add a processing filter')

    # The icon of the displayed objects
    icon_name = Str('filter.ico')

    # A reference to the registry, to generate this list.
    items_list_source = registry.filters


###############################################################################
# ModuleFilterAdderNode class
###############################################################################
class ModuleFilterAdderNode(AdderNode):
    """ Tree node that presents a view to the user to add filter and
        modules.
    """

    # The string to display on the icon in the TreeEditor.
    label = 'Add module or filter'

    # An adder node for modules
    modules = Instance(ModuleAdderNode, ())

    # An adder node for filters
    filters = Instance(FilterAdderNode, ())

    def _object_changed(self):
        """ Propagate the object to the sub nodes.
        """
        self.filters.object = self.object
        self.modules.object = self.object

    # Trait view to show in the Mayavi current object panel.
    view = View(
                Group(Item('modules', style='custom', springy=True,
                            resizable=True,
                            height=1.,
                            ),
                    show_labels=False,
                    label='Visualization modules'),
                Group(Item('filters', style='custom', springy=True,
                            resizable=True,
                            height=1.,
                            ),
                    show_labels=False,
                    label='Processing filters'),
                )


### EOF #######################################################################



# Author: Suyog Dutt Jain <suyog.jain@aero.iitb.ac.in>
#         Prabhu Ramachandran <prabhu_r@users.sf.net>
# Copyright (c) 2008-2020,  Enthought, Inc.
# License: BSD Style.

# Standard library imports.
from os.path import abspath
from io import BytesIO
import platform
import copy
import numpy
import unittest

# Local imports.
from mayavi.core.null_engine import NullEngine

# Enthought library imports
from mayavi.sources.array_source import ArraySource
from mayavi.modules.outline import Outline
from mayavi.modules.glyph import Glyph
from mayavi.modules.vector_cut_plane import VectorCutPlane


class TestGlyph(unittest.TestCase):

    def make_data(self):
        """Trivial data -- creates an elementatry scalar field and a
        constant vector field along the 'x' axis."""
        s = numpy.arange(0.0, 10.0, 0.01)
        s = numpy.reshape(s, (10, 10, 10))
        s = numpy.transpose(s)

        v = numpy.zeros(3000, 'd')
        v[1::3] = 1.0
        v = numpy.reshape(v, (10, 10, 10, 3))
        return s, v

    def setUp(self):
        e = NullEngine()
        # Uncomment to see visualization for debugging etc.
        # e = Engine()
        e.start()
        s = e.new_scene()
        self.e = e
        self.s = s

        ############################################################
        # Create a new scene and set up the visualization.

        d = ArraySource()
        sc, vec = self.make_data()
        d.origin = (-5, -5, -5)
        d.scalar_data = sc
        d.vector_data = vec

        e.add_source(d)

        # Create an outline for the data.
        o = Outline()
        e.add_module(o)
        # Glyphs for the scalars
        g = Glyph()
        e.add_module(g)
        g.glyph.glyph_source.glyph_position = 'center'
        g.glyph.glyph.vector_mode = 'use_normal'
        g.glyph.glyph.scale_factor = 0.5
        g.glyph.mask_points.on_ratio = 20
        g.actor.property.line_width = 1.0

        v = VectorCutPlane()
        glyph = v.glyph
        gs = glyph.glyph_source
        gs.glyph_position = 'tail'
        gs.glyph_source = gs.glyph_list[1]
        e.add_module(v)
        v.implicit_plane.trait_set(normal=(0, 1, 0), origin=(0, 3, 0))

        v = VectorCutPlane()
        glyph = v.glyph
        gs = glyph.glyph_source
        gs.glyph_source = gs.glyph_list[2]
        gs.glyph_position = 'head'
        e.add_module(v)
        v.implicit_plane.trait_set(normal=(0, 1, 0), origin=(0, -2, 0))
        self.g=g
        self.v=v
        self.scene = e.current_scene
        return

    def tearDown(self):
        """For necessary clean up, automatically called by TestCase after the test methods have been invoked"""
        self.e.stop()
        return

    def check(self, mask=False, mask_random_mode=False):
        """Do the actual testing with and without masking. For masking,
           both the presence and absence of random mode is also tested.
        """
        s = self.scene
        src = s.children[0]
        g = src.children[0].children[1]
        self.assertEqual(g.glyph.glyph_source.glyph_position, 'center')
        self.assertEqual(g.glyph.glyph.vector_mode, 'use_normal')
        self.assertEqual(g.glyph.glyph.scale_factor, 0.5)
        self.assertEqual(g.actor.property.line_width, 1.0)
        # Test masking
        n_output_points = src.outputs[0].output.number_of_points
        n_glyph_input_points = g.glyph.glyph.input.number_of_points
        if mask:
            self.assertNotEqual(n_glyph_input_points, 0)
            if mask_random_mode:
                self.assertLessEqual(n_glyph_input_points, n_output_points)
            else:
                on_ratio = g.glyph.mask_points.on_ratio
                self.assertEqual(n_glyph_input_points,
                                 n_output_points / on_ratio)
        else:
            self.assertEqual(n_glyph_input_points, n_output_points)

        v = src.children[0].children[2]
        glyph = v.glyph
        gs = glyph.glyph_source
        self.assertEqual(gs.glyph_position, 'tail')
        self.assertEqual(gs.glyph_source, gs.glyph_list[1])
        self.assertEqual(numpy.allclose(v.implicit_plane.normal,
                                        (0., 1., 0.)), True)

        v = src.children[0].children[3]
        glyph = v.glyph
        gs = glyph.glyph_source
        self.assertEqual(gs.glyph_source, gs.glyph_list[2])
        self.assertEqual(gs.glyph_position, 'head')
        self.assertEqual(numpy.allclose(v.implicit_plane.normal,
                                        (0., 1., 0.)), True)

    def test_glyph(self):
        "Test if the test fixture works"
        self.check()

    @unittest.skipIf(platform.system() == "Darwin", "Test crashes on OSX. See issue #373")
    def test_mask_input_points_with_random_mode(self):
        """Test if masking input points works with random mode.
           Tests Issue #165"""
        s = self.scene
        src = s.children[0]
        g = src.children[0].children[1]
        g.glyph.mask_input_points = True
        self.check(mask=True, mask_random_mode=True)

    def test_mask_input_points_without_random_mode(self):
        """Test if masking input points works without random mode.
           Tests Issue #165"""
        s = self.scene
        src = s.children[0]
        g = src.children[0].children[1]
        g.glyph.mask_points.random_mode = 0
        g.glyph.mask_input_points = True
        self.check(mask=True)

    def test_components_changed(self):
        """"Test if the modules respond correctly when the components
            are changed."""

        g = self.g
        v = self.v
        g.actor = g.actor.__class__()
        glyph = g.glyph
        g.glyph = glyph.__class__()
        g.glyph = glyph

        glyph = v.glyph
        v.glyph = glyph.__class__()
        v.glyph = glyph
        v.actor = v.actor.__class__()
        v.cutter = v.cutter.__class__()
        ip = v.implicit_plane
        v.implicit_plane = ip.__class__()
        v.implicit_plane = ip
        self.check()

    def test_save_and_restore(self):
        """Test if saving a visualization and restoring it works."""
        engine = self.e
        scene = self.scene
        # Save visualization.
        f = BytesIO()
        f.name = abspath('test.mv2')  # We simulate a file.
        engine.save_visualization(f)
        f.seek(0)  # So we can read this saved data.

        # Remove existing scene.

        engine.close_scene(scene)

        # Load visualization
        engine.load_visualization(f)
        self.scene = engine.current_scene

        self.check()

    def test_deepcopied(self):
        """Test if the MayaVi2 visualization can be deep-copied."""
        ############################################################
        # Test if the MayaVi2 visualization can be deep-copied.

        # Pop the source object.
        s =  self.scene
        sources = s.children
        s.children = []
        # Add it back to see if that works without error.
        s.children.extend(sources)

        self.check()

        # Now deepcopy the source and replace the existing one with
        # the copy.  This basically simulates cutting/copying the
        # object from the UI via the right-click menu on the tree
        # view, and pasting the copy back.
        sources1 = copy.deepcopy(sources)
        s.children[:] = sources1
        self.check()

if __name__ == '__main__':
    unittest.main()



"""Functionality to display Mayavi scenes inside Jupyter notebooks.
"""
import base64
from itertools import count

from tvtk.api import tvtk
from tvtk.common import configure_input


_backend = 'ipy'
_width = None
_height = None
_local = True
_widget_manager = None
_counter = count()


def init(backend='ipy', width=None, height=None, local=True):
    """Initialize a suitable backend for Jupyter notebooks.

    **Parameters**

    backend :str: one of ('png', 'x3d', 'ipy')
    width :int: suggested default width of the element
    height :int: suggested default height of the element
    local :bool: Use local copy of x3dom.js instead of online version.
    """
    global _backend, _width, _height, _local, _widget_manager
    backends = ('png', 'x3d', 'ipy')
    error_msg = "Backend must be one of %r, got %s" % (backends, backend)
    assert backend in backends, error_msg
    from mayavi import mlab
    mlab.options.offscreen = True
    _backend = backend
    _width, _height = width, height
    _local = local
    if backend == 'ipy':
        _setup_widget_manager()
    _monkey_patch_for_ipython()
    print("Notebook initialized with %s backend." % backend)


def _setup_widget_manager():
    global _widget_manager
    if _widget_manager is None:
        from mayavi.tools.remote.ipy_remote import WidgetManager
        _widget_manager = WidgetManager()


def _monkey_patch_for_ipython():
    from mayavi.core.base import Base
    from tvtk.pyface.tvtk_scene import TVTKScene
    Base._ipython_display_ = _ipython_display_
    TVTKScene._ipython_display_ = _ipython_display_


def _ipython_display_(self):
    """Method for displaying elements on the Jupyter notebook.
    """
    from IPython.display import HTML
    from IPython.display import display as idisplay

    if hasattr(self, 'render_window'):
        scene = self
    elif hasattr(self, 'scene'):
        scene = self.scene
    if _backend == 'png':
        return idisplay(HTML(scene_to_png(scene)))
    elif _backend == 'x3d':
        return idisplay(HTML(scene_to_x3d(scene)))
    elif _backend == 'ipy':
        return idisplay(scene_to_ipy(scene))


def _fix_x3d_header(x3d):
    id = 'scene_%d' % next(_counter)
    rep = '<X3D profile="Immersive" version="3.0" id="%s" ' % id
    if _width is not None:
        rep += 'width="%dpx" ' % _width
    if _height is not None:
        rep += 'height="%dpx" ' % _height
    rep += '>'
    if isinstance(x3d, bytes):
        x3d = x3d.decode("utf-8", "ignore") 
    x3d = x3d.replace(
        '<X3D profile="Immersive" version="3.0">',
        rep
    )
    return x3d


def scene_to_x3d(scene):
    ex = tvtk.X3DExporter()
    ex.input = scene.render_window
    lm = scene.light_manager.light_mode
    # The default raymond mode is too bright so switch back to vtk mode.
    scene.light_manager.light_mode = 'vtk'
    ex.write_to_output_string = True
    ex.update()
    ex.write()
    # Switch back
    scene.light_manager.light_mode = lm
    if _local:
        url_base = "nbextensions/mayavi/x3d"
    else:
        url_base = "https://www.x3dom.org/download/1.7.2"
    x3d_elem = _fix_x3d_header(ex.output_string)
    html = '''
    %s
    <script type="text/javascript">
    require(["%s/x3dom"], function(x3dom) {
        var x3dom_css = document.getElementById("x3dom-css");
        if (x3dom_css === null) {
            var l = document.createElement("link");
            l.setAttribute("rel", "stylesheet");
            l.setAttribute("type", "text/css");
            l.setAttribute("href", require.toUrl("%s/x3dom.css"));
            l.setAttribute("id", "x3dom-css");
            $("head").append(l);
        }
        if (typeof x3dom != 'undefined') {
            x3dom.reload();
        }
        else if (typeof window.x3dom != 'undefined') {
            window.x3dom.reload();
        }
    })
    </script>
    ''' % (x3d_elem, url_base, url_base)
    return html


def scene_to_png(scene):
    w2if = tvtk.WindowToImageFilter()
    w2if.input = scene.render_window
    ex = tvtk.PNGWriter()
    ex.write_to_memory = True
    configure_input(ex, w2if)
    ex.update()
    ex.write()
    data = base64.b64encode(ex.result.to_array()).decode('ascii')
    html = '<img src="data:image/png;base64,%s" alt="PNG image"></img>'
    return html % data


def scene_to_ipy(scene):
    return _widget_manager.scene_to_ipy(scene)


def display(obj, backend=None):
    """Display given object on Jupyter notebook using given backend.

    This is largely for testing.
    """
    global _backend
    backend = _backend if backend is None else backend
    if backend == 'ipy':
        _setup_widget_manager()
    orig_backend = _backend
    _backend = backend
    result = _ipython_display_(obj)
    _backend = orig_backend
    return result



"""
Test for MlabSource and its subclasses.
"""
# Author: Prabhu Ramachandran <prabhu@aero.iitb.ac.in>
# Copyright (c) 2008, Enthought, Inc.
# License: BSD Style.

import unittest
import numpy as np
from unittest.mock import patch

from tvtk.api import tvtk
from mayavi.tools import sources
from mayavi.sources.vtk_data_source import VTKDataSource


###############################################################################
# `TestMGlyphSource`
###############################################################################
class TestMGlyphSource(unittest.TestCase):
    def setUp(self):
        self.x = x = np.ones(10, float)
        self.y = y = np.ones(10, float)*2.0
        self.z = z = np.linspace(0, 10, 10)
        self.v = v = np.ones((10, 3), float)*10.0
        self.s = s = np.ones(10, float)
        src = sources.MGlyphSource()
        src.reset(x=x, y=y, z=z, u=v[:, 0], v=v[:, 1], w=v[:, 2], scalars=s)
        self.src = src

    def tearDown(self):
        return

    def get_data(self):
        return self.x, self.y, self.z, self.v, self.s, self.src

    def check_traits(self):
        """Check if the sources traits are set correctly."""
        x, y, z, v, s, src = self.get_data()
        # Check if points are set correctly.
        self.assertEqual(np.alltrue(src.points[:, 0].ravel() == x.ravel()),
                         True)
        self.assertEqual(np.alltrue(src.points[:, 1].ravel() == y.ravel()),
                         True)
        self.assertEqual(np.alltrue(src.points[:, 2].ravel() == z.ravel()),
                         True)
        # Check the vectors and scalars.
        self.assertEqual(np.alltrue(src.vectors == v), True)
        self.assertEqual(np.alltrue(src.scalars == s), True)

    def check_dataset(self):
        """Check the TVTK dataset."""
        x, y, z, v, s, src = self.get_data()
        # Check if the dataset is setup right.
        pts = src.dataset.points.to_array()
        self.assertEqual(np.alltrue(pts[:, 0].ravel() == x.ravel()), True)
        self.assertEqual(np.alltrue(pts[:, 1].ravel() == y.ravel()), True)
        self.assertEqual(np.alltrue(pts[:, 2].ravel() == z.ravel()), True)
        vec = src.dataset.point_data.vectors.to_array()
        sc = src.dataset.point_data.scalars.to_array()
        self.assertEqual(np.alltrue(vec == v), True)
        self.assertEqual(np.alltrue(sc == s.ravel()), True)

    def test_geometry_filter_works(self):
        # This tests the bug #1071 and segfaults without the fix on VTK 9.x.
        f = tvtk.GeometryFilter(extent_clipping=True)
        f.set_input_data_object(self.src.dataset)
        f.update()

    def test_reset_with_same_size_data(self):
        x, y, z, v, s, src = self.get_data()
        self.check_traits()
        self.check_dataset()

        # Call reset again with just a few things changed to see if it
        # works correctly.
        x *= 5
        s *= 10
        v *= 0.1
        src.reset(x=x, u=v[:, 0], v=v[:, 1], w=v[:, 2], scalars=s)

        self.check_traits()
        self.check_dataset()

    def test_reset_with_resized_data(self):
        x, y, z, v, s, src = self.get_data()
        self.check_traits()
        self.check_dataset()
        # Call reset again with just a few things changed to see if it
        # works correctly.

        self.x = x = np.ones(20, float)*30.0
        self.y = y = np.ones(20, float)*30.0
        self.z = z = np.ones(20, float)*30.0
        self.s = s = np.ones(20, float)
        self.v = v = np.ones((20, 3), float)*30.0

        src.reset(x=x, y=y, z=z, u=v[:, 0], v=v[:, 1], w=v[:, 2],
                  scalars=s)
        self.check_traits()
        self.check_dataset()

    def test_reset_strange_shape(self):
        " Test the reset method when the inputs are 2-d arrays."

        x, y, z, v, s, src = self.get_data()
        self.x = x = np.reshape(x, (5, 2))
        self.y = y = np.reshape(y, (5, 2))
        self.z = z = np.reshape(z, (5, 2))
        u = np.reshape(v[:, 0], (5, 2))
        vv = np.reshape(v[:, 1], (5, 2))
        w = np.reshape(v[:, 2], (5, 2))
        self.s = s = np.reshape(s, (5, 2))
        src.reset(x=x, y=y, z=z, u=u, v=vv, w=w, scalars=s)
        self.check_traits()
        self.check_dataset()

    def test_handlers(self):
        "Test if the various static handlers work correctly."
        x, y, z, v, s, src = self.get_data()
        x *= 2
        y *= 2
        z *= 2
        v *= 2
        s *= 2
        src.x = x
        src.y = y
        src.z = z
        src.u = v[:, 0]
        src.v = v[:, 1]
        src.w = v[:, 2]
        src.scalars = s
        self.check_traits()
        self.check_dataset()

    def test_handlers_strange_shape(self):
        """Test if the various static handlers work correctly for strange shapes.
        """
        # Initialize with 2-d array data.
        x, y, z, v, s, src = self.get_data()
        x = np.reshape(x, (5, 2))
        y = np.reshape(y, (5, 2))
        z = np.reshape(z, (5, 2))
        u = np.reshape(v[:, 0], (5, 2))
        vv = np.reshape(v[:, 1], (5, 2))
        w = np.reshape(v[:, 2], (5, 2))
        s = np.reshape(s, (5, 2))
        src.reset(x=x, y=y, z=z, u=u, v=vv, w=w, scalars=s)

        # modify variables in src to check handlers
        self.x = src.x = 2*x
        self.y = src.y = 2*y
        self.z = src.z = 2*z
        src.u = 2*z
        src.v = 2*vv
        src.w = 2*w
        self.v[:, 0] = src.u.ravel()
        self.v[:, 1] = src.v.ravel()
        self.v[:, 2] = src.w.ravel()
        self.s = src.scalars = 2*s
        self.check_traits()
        self.check_dataset()

    def test_set(self):
        "Test if the set method works correctly."
        x, y, z, v, s, src = self.get_data()
        x *= 2
        z *= 2
        s *= 2
        src.trait_set(x=x, z=z, scalars=s)
        self.check_traits()
        self.check_dataset()

    def test_strange_shape(self):
        " Test the MGlyphSource with strange shapes for the arguments "
        x, y, z, v, s, src = self.get_data()
        x = y = z = v = s = 0
        src.reset(x=x, y=y, z=z, u=v, v=v, w=v, scalars=None)
        src.reset(x=x, y=y, z=z, u=v, v=v, w=v, scalars=s)
        x = y = z = v = s = 1
        src.trait_set(x=x, y=y, z=z, u=v, v=v, w=v, scalars=None)
        src.trait_set(x=x, y=y, z=z, u=v, v=v, w=v, scalars=s)

    @patch('mayavi.tools.engine_manager.options.backend', 'test')
    def test_reset_changes_pipeline(self):
        # Given
        from mayavi import mlab
        x, y, z = np.random.random((3, 10))
        g = mlab.points3d(x, y, z, x*x + y*y + z*z)
        bounds = g.actor.actor.bounds

        # When
        x, y, z = np.random.random((3, 20))
        g.mlab_source.reset(x=x, y=y, z=z, scalars=x*x + y*y + z*z)

        # Then
        self.assertFalse(np.allclose(bounds, g.actor.actor.bounds))


class TestMVerticalSource(unittest.TestCase):
    def setUp(self):
        self.x = x = np.ones(10, float)
        self.y = y = np.ones(10, float)*2.0
        self.z = z = np.linspace(0, 10, 10)
        self.s = s = np.ones(10, float)
        src = sources.MVerticalGlyphSource()
        src.reset(x=x, y=y, z=z, scalars=s)
        self.src = src

    def tearDown(self):
        return

    def get_data(self):
        return self.x, self.y, self.z, self.s, self.src

    def check_traits(self):
        """Check if the sources traits are set correctly."""
        x, y, z, s, src = self.get_data()
        # Check if points are set correctly.
        self.assertEqual(np.alltrue(src.points[:, 0].ravel() == x), True)
        self.assertEqual(np.alltrue(src.points[:, 1].ravel() == y), True)
        self.assertEqual(np.alltrue(src.points[:, 2].ravel() == z), True)
        # Check the vectors and scalars.
        self.assertEqual(np.alltrue(src.vectors[:, -1] == s), True)
        self.assertEqual(np.alltrue(src.vectors[:, :-1] == 1), True)
        self.assertEqual(np.alltrue(src.scalars == s), True)

    def check_dataset(self):
        """Check the TVTK dataset."""
        x, y, z, s, src = self.get_data()
        # Check if the dataset is setup right.
        pts = src.dataset.points.to_array()
        self.assertEqual(np.alltrue(pts[:, 0].ravel() == x), True)
        self.assertEqual(np.alltrue(pts[:, 1].ravel() == y), True)
        self.assertEqual(np.alltrue(pts[:, 2].ravel() == z), True)
        vec = src.dataset.point_data.vectors.to_array()
        sc = src.dataset.point_data.scalars.to_array()
        self.assertEqual(np.alltrue(vec[:, -1] == s), True)
        self.assertEqual(np.alltrue(vec[:, :-1] == 1), True)
        self.assertEqual(np.alltrue(sc == s), True)

    def test_reset(self):
        "Test the reset method."
        x, y, z, s, src = self.get_data()
        self.check_traits()
        self.check_dataset()

        # Call reset again with just a few things changed to see if it
        # works correctly.
        x *= 5
        s *= 10
        src.reset(x=x, scalars=s)

        self.check_traits()
        self.check_dataset()

    def test_reset1(self):

        "Test the reset method."
        x, y, z, s, src = self.get_data()
        self.check_traits()
        self.check_dataset()
        # Call reset again with just a few things changed to see if it
        # works correctly.

        self.x = x = np.ones(20, float)*30.0
        self.y = y = np.ones(20, float)*30.0
        self.z = z = np.ones(20, float)*30.0
        points = np.ones((20, 3), float)*30.0
        self.s = s = np.ones(20, float)

        src.reset(x=x, y=y, z=z, scalars=s, points=points)
        self.check_traits()
        self.check_dataset()

    def test_handlers(self):
        "Test if the various static handlers work correctly."
        x, y, z, s, src = self.get_data()
        x *= 2
        y *= 2
        z *= 2
        s *= 2
        src.x = x
        src.y = y
        src.z = z
        src.scalars = s
        self.check_traits()
        self.check_dataset()

    def test_set(self):
        "Test if the set method works correctly."
        x, y, z, s, src = self.get_data()
        x *= 2
        z *= 2
        s *= 2
        src.trait_set(x=x, z=z, scalars=s)
        self.check_traits()
        self.check_dataset()


class TestMArraySource(unittest.TestCase):
    def setUp(self):
        x, y, z = np.ogrid[-10:10:11j, -10:10:12j, -10:10:20j]
        self.x, self.y, self.z = x, y, z
        dims = (x.shape[0], y.shape[1], z.shape[2])
        self.v = v = np.ones(dims + (3, ), float)
        v[..., 2] = 2
        v[..., 2] = 3
        self.s = s = np.ones(dims, float)
        src = sources.MArraySource()
        src.reset(
            x=x, y=y, z=z, u=v[..., 0], v=v[..., 1], w=v[..., 2], scalars=s
        )
        self.src = src

    def tearDown(self):
        return

    def get_data(self):
        return self.x, self.y, self.z, self.v, self.s, self.src

    def check_traits(self):
        """Check if the sources traits are set correctly."""
        x, y, z, v, s, src = self.get_data()
        # Check if points are set correctly.
        self.assertEqual(np.alltrue(src.x == x), True)
        self.assertEqual(np.alltrue(src.y == y), True)
        self.assertEqual(np.alltrue(src.z == z), True)
        # Check the vectors and scalars.
        self.assertEqual(np.alltrue(src.vectors == v), True)
        self.assertEqual(np.alltrue(src.scalars == s), True)

    def check_dataset(self):
        """Check the TVTK dataset."""
        x, y, z, v, s, src = self.get_data()
        # Check if the dataset is setup right.
        dx = x[1, 0, 0] - x[0, 0, 0]
        dy = y[0, 1, 0] - y[0, 0, 0]
        dz = z[0, 0, 1] - z[0, 0, 0]
        origin = [x.min(), y.min(), z.min()]
        spacing = [dx, dy, dz]
        dimensions = (x.shape[0], y.shape[1], z.shape[2])
        ds = src.dataset
        self.assertEqual(np.all(src.m_data.origin == origin), True)
        self.assertEqual(np.allclose(src.m_data.spacing, spacing), True)
        self.assertEqual(np.allclose(ds.dimensions, dimensions), True)

        vec = src.dataset.point_data.vectors.to_array()
        sc = src.dataset.point_data.scalars.to_array()
        v1 = v.transpose((2, 0, 1, 3))
        self.assertEqual(np.alltrue(vec.ravel() == v1.ravel()), True)
        s1 = s.transpose()
        self.assertEqual(np.alltrue(sc.ravel() == s1.ravel()), True)

    def test_reset(self):
        "Test the reset method."
        x, y, z, v, s, src = self.get_data()
        self.check_traits()
        self.check_dataset()

        # Call reset again with just a few things changed to see if it
        # works correctly.
        x *= 5
        s *= 10
        v *= 0.1
        src.reset(x=x, u=v[..., 0], v=v[..., 1], w=v[..., 2], scalars=s)

        self.check_traits()
        self.check_dataset()

    def test_reset1(self):
        "Test the reset method."
        x, y, z, v, s, src = self.get_data()
        self.check_traits()
        self.check_dataset()
        # Call reset again with just a few things changed to see if it
        # works correctly.

        x, y, z = np.ogrid[-10:10:11j, -10:10:12j, -10:10:20j]
        self.x, self.y, self.z = x, y, z

        dims = (x.shape[0], y.shape[1], z.shape[2])
        self.v = v = np.ones(dims + (3, ), float)
        v[..., 2] = 2
        v[..., 2] = 3
        self.s = s = np.ones(dims, float)
        src = sources.MArraySource()
        src.reset(
            x=x,
            y=y,
            z=z,
            u=v[..., 0],
            v=v[..., 1],
            w=v[..., 2],
            scalars=s,
            vectors=v
        )
        self.check_traits()
        self.check_dataset()

    def test_handlers(self):
        "Test if the various static handlers work correctly."
        x, y, z, v, s, src = self.get_data()
        x *= 2
        y *= 2
        z *= 2
        v *= 2
        s *= 2
        src.x = x
        src.y = y
        src.z = z
        src.u = v[..., 0]
        src.v = v[..., 1]
        src.w = v[..., 2]
        src.scalars = s
        self.check_traits()
        self.check_dataset()

    def test_set(self):
        "Test if the set method works correctly."
        x, y, z, v, s, src = self.get_data()
        x *= 2
        z *= 2
        s *= 2
        src.trait_set(x=x, z=z, scalars=s)
        self.check_traits()
        self.check_dataset()


class TestMLineSource(unittest.TestCase):
    def setUp(self):
        self.x = x = np.ones(10, float)
        self.y = y = np.ones(10, float) * 2.0
        self.z = z = np.linspace(0, 10, 10)
        self.s = s = np.ones(10, float)
        src = sources.MLineSource()
        src.reset(x=x, y=y, z=z, scalars=s)
        self.src = src

    def tearDown(self):
        return

    def get_data(self):
        return self.x, self.y, self.z, self.s, self.src

    def check_traits(self):
        """Check if the sources traits are set correctly."""
        x, y, z, s, src = self.get_data()
        # Check if points are set correctly.
        self.assertEqual(np.alltrue(src.points[:, 0].ravel() == x), True)
        self.assertEqual(np.alltrue(src.points[:, 1].ravel() == y), True)
        self.assertEqual(np.alltrue(src.points[:, 2].ravel() == z), True)
        # Check the scalars.
        self.assertEqual(np.alltrue(src.scalars == s), True)

    def check_dataset(self):
        """Check the TVTK dataset."""
        x, y, z, s, src = self.get_data()
        # Check if the dataset is setup right.
        pts = src.dataset.points.to_array()
        self.assertEqual(np.alltrue(pts[:, 0].ravel() == x), True)
        self.assertEqual(np.alltrue(pts[:, 1].ravel() == y), True)
        self.assertEqual(np.alltrue(pts[:, 2].ravel() == z), True)
        sc = src.dataset.point_data.scalars.to_array()
        self.assertEqual(np.alltrue(sc == s), True)

    def test_reset(self):
        "Test the reset method."
        x, y, z, s, src = self.get_data()
        self.check_traits()
        self.check_dataset()

        # Call reset again with just a few things changed to see if it
        # works correctly.
        x *= 5
        s *= 10
        src.reset(x=x, scalars=s)

        self.check_traits()
        self.check_dataset()

        y *= 6
        x *= 4
        src.reset(x=x, y=y)

        self.check_traits()
        self.check_dataset()

        s *= 4.5
        y /= 4
        src.reset(y=y, s=s)

        self.check_traits()
        self.check_dataset()

    def test_reset1(self):

        "Test the reset method."
        x, y, z, s, src = self.get_data()
        self.check_traits()
        self.check_dataset()
        # Call reset again with just a few things changed to see if it
        # works correctly.

        self.x = x = np.ones(20, float)*30.0
        self.y = y = np.ones(20, float)*30.0
        self.z = z = np.ones(20, float)*30.0
        points = np.ones((20, 3), float)*30.0
        self.s = s = np.ones(20, float)
        src.reset(x=x, y=y, z=z, scalars=s, points=points)
        self.check_traits()
        self.check_dataset()

    def test_handlers(self):
        "Test if the various static handlers work correctly."
        x, y, z, s, src = self.get_data()
        x *= 2
        y *= 2
        z *= 2
        s *= 2
        src.x = x
        src.y = y
        src.z = z
        src.scalars = s
        self.check_traits()
        self.check_dataset()

    def test_set(self):
        "Test if the set method works correctly."
        x, y, z, s, src = self.get_data()
        x *= 2
        z *= 2
        s *= 2
        src.trait_set(x=x, z=z, scalars=s)
        self.check_traits()
        self.check_dataset()

        y *= 2
        s *= 2
        src.trait_set(y=y, scalars=s)
        self.check_traits()
        self.check_dataset()

    @patch('mayavi.tools.engine_manager.options.backend', 'test')
    def test_reset_changes_pipeline(self):
        # Given
        from mayavi import mlab
        x = np.linspace(0, 1, 10)
        lines = mlab.plot3d(x, x, x, x)
        bounds = lines.actor.actor.bounds

        # When
        x = np.linspace(0, 2, 20)
        lines.mlab_source.reset(x=x, y=x, z=x, scalars=x)

        # Then
        self.assertFalse(np.allclose(bounds, lines.actor.actor.bounds))

    def test_set_without_scalars_works(self):
        # Given
        x, y, z, s, src = self.get_data()
        src = sources.MLineSource()
        src.reset(x=x, y=y, z=z)

        # When
        src.m_data = VTKDataSource(data=src.dataset)
        src.set(y=y+1)

        # Then
        self.assertTrue(np.allclose(src.y, y + 1))

        # When
        src.reset(x=x, y=y+1, z=z+1)

        # Then
        self.assertTrue(np.allclose(src.y, y + 1))
        self.assertTrue(np.allclose(src.z, z + 1))

    def test_set_without_scalars_attribute_works(self):
        # Given
        class MySource(sources.MlabSource):
            x = sources.ArrayNumberOrNone

        src = MySource(x=1.0, dataset=tvtk.PolyData())
        src.m_data = VTKDataSource(data=src.dataset)

        # When
        src.update()

        # Then
        np.testing.assert_almost_equal(src.x, 1.0)


class TestMArray2DSource(unittest.TestCase):
    def setUp(self):
        x, y = np.mgrid[-10:10:11j, -10:10:12j]

        self.x, self.y = x, y
        dims = (x.shape[0], y.shape[1])
        self.s = s = np.ones(dims, float)
        src = sources.MArray2DSource()
        src.reset(x=x, y=y, scalars=s)
        self.src = src

    def tearDown(self):
        return

    def get_data(self):
        return self.x, self.y, self.s, self.src

    def check_traits(self):
        """Check if the sources traits are set correctly."""
        x, y, s, src = self.get_data()

        # Check if points are set correctly.
        self.assertEqual(np.alltrue(src.x == x), True)
        self.assertEqual(np.alltrue(src.y == y), True)
        # Check the scalars.
        self.assertEqual(np.alltrue(src.scalars == s), True)

    def check_dataset(self):
        """Check the TVTK dataset."""
        x, y, s, src = self.get_data()
        # Check if the dataset is setup right.
        x = np.atleast_2d(x.squeeze().T)[0, :].squeeze()
        y = np.atleast_2d(y.squeeze())[0, :].squeeze()
        dx = x[1] - x[0]
        dy = y[1] - y[0]

        origin = [x.min(), y.min(), 0]
        spacing = [dx, dy, 1]
        ds = src.dataset
        self.assertEqual(np.all(ds.origin == origin), True)
        self.assertEqual(np.allclose(src.m_data.spacing, spacing), True)

        sc = src.dataset.point_data.scalars.to_array()
        s1 = s.transpose()
        self.assertEqual(np.alltrue(sc.ravel() == s1.ravel()), True)

    def test_reset(self):
        x, y, s, src = self.get_data()

        self.check_traits()
        self.check_dataset()

        # Call reset again with just a few things changed to see if it
        # works correctly.
        x *= 5
        s *= 10
        src.reset(x=x, y=y, scalars=s)

        self.check_traits()
        self.check_dataset()

    def test_handlers(self):
        "Test if the various static handlers work correctly."
        x, y, s, src = self.get_data()
        x *= 2
        y *= 2
        s *= 2
        src.x = x
        src.y = y
        src.scalars = s

        self.check_traits()
        self.check_dataset()

    def test_set(self):
        "Test if the set method works correctly."
        x, y, s, src = self.get_data()
        x *= 2
        s *= 2
        src.trait_set(x=x, scalars=s)

        self.check_traits()
        self.check_dataset()

        y *= 9
        s *= 2
        src.trait_set(y=y, scalars=s)
        self.check_traits()
        self.check_dataset()


class TestMGridSource(unittest.TestCase):
    def setUp(self):
        self.x = x = np.ones([10, 10], float)
        self.y = y = np.ones([10, 10], float) * 2.0
        self.z = z = np.ones([10, 10], float) * 3.0
        self.s = s = np.ones([10, 10], float)
        src = sources.MGridSource()
        src.reset(x=x, y=y, z=z, scalars=s)
        self.src = src

    def tearDown(self):
        return

    def get_data(self):
        return self.x, self.y, self.z, self.s, self.src

    def check_traits(self):
        """Check if the sources traits are set correctly."""
        x, y, z, s, src = self.get_data()

        # Check if points are set correctly.
        self.assertEqual(
            np.alltrue(src.points[:, 0].ravel() == x.ravel()), True
        )
        self.assertEqual(
            np.alltrue(src.points[:, 1].ravel() == y.ravel()), True
        )
        self.assertEqual(
            np.alltrue(src.points[:, 2].ravel() == z.ravel()), True
        )
        # Check the  scalars.

        self.assertEqual(np.alltrue(src.scalars == s), True)

    def check_dataset(self):
        """Check the TVTK dataset."""
        x, y, z, s, src = self.get_data()
        # Check if the dataset is setup right.

        pts = src.dataset.points.to_array()
        self.assertEqual(np.alltrue(pts[:, 0].ravel() == x.ravel()), True)
        self.assertEqual(np.alltrue(pts[:, 1].ravel() == y.ravel()), True)
        self.assertEqual(np.alltrue(pts[:, 2].ravel() == z.ravel()), True)
        sc = src.dataset.point_data.scalars.to_array()
        self.assertEqual(np.alltrue(sc == s.ravel()), True)

    def test_reset(self):
        "Test the reset method."

        x, y, z, s, src = self.get_data()
        self.check_traits()
        self.check_dataset()

        # Call reset again with just a few things changed to see if it
        # works correctly.
        x *= 5
        s *= 10

        src.reset(x=x, scalars=s)
        self.check_traits()
        self.check_dataset()

    def test_handlers(self):
        "Test if the various static handlers work correctly."
        x, y, z, s, src = self.get_data()
        x *= 2
        y *= 2
        z *= 2
        s *= 2
        src.x = x
        src.y = y
        src.z = z
        src.scalars = s
        self.check_traits()
        self.check_dataset()

    def test_set(self):
        "Test if the set method works correctly."
        x, y, z, s, src = self.get_data()
        x *= 2
        z *= 2
        s *= 2
        src.trait_set(x=x, z=z, scalars=s)
        self.check_traits()
        self.check_dataset()

    @patch('mayavi.tools.engine_manager.options.backend', 'test')
    def test_reset_changes_pipeline(self):
        # Given
        from mayavi import mlab
        s = slice(0, 1, 10j)
        x, y = np.mgrid[s, s]
        obj = mlab.mesh(x, y, x*y, scalars=x)
        bounds = obj.actor.actor.bounds

        # When
        s = slice(0, 5, 20j)
        x, y = np.mgrid[s, s]
        obj.mlab_source.reset(x=x, y=y, z=x*y, scalars=x)

        # Then
        self.assertFalse(np.allclose(bounds, obj.actor.actor.bounds))


class TestMArray2DSourceNoArgs(unittest.TestCase):
    """Special Test Case for MArray2DSource when both x and y are specified as
    None"""

    def setUp(self):

        x = None
        y = None

        self.x, self.y = x, y

        if x is not None and y is not None:
            dims = (x.shape[0], y.shape[1])
        else:
            dims = (10, 10)

        self.s = s = np.ones(dims, float)
        src = sources.MArray2DSource()
        src.reset(x=x, y=y, scalars=s)
        self.src = src

    def tearDown(self):
        return

    def get_data(self):
        return self.x, self.y, self.s, self.src

    def check_traits(self):
        """Check if the sources traits are set correctly."""
        x, y, s, src = self.get_data()
        # Check if points are set correctly.

        if x is not None and y is not None:
            self.assertEqual(np.alltrue(src.x == x), True)
            self.assertEqual(np.alltrue(src.y == y), True)

        else:
            nx, ny = s.shape
            x1, y1 = np.mgrid[-nx/2.:nx/2, -ny/2.:ny/2]
            self.assertEqual(np.alltrue(src.x == x1), True)
            self.assertEqual(np.alltrue(src.y == y1), True)

        # Check the scalars.
        self.assertEqual(np.alltrue(src.scalars == s), True)

    def check_dataset(self):
        """Check the TVTK dataset."""
        x, y, s, src = self.get_data()
        # Check if the dataset is setup right.

        nx, ny = src.scalars.shape

        if x is None and y is None:
            x, y = np.mgrid[-nx/2.:nx/2, -ny/2.:ny/2]

        x = np.atleast_2d(x.squeeze().T)[0, :].squeeze()
        y = np.atleast_2d(y.squeeze())[0, :].squeeze()
        dx = x[1] - x[0]
        dy = y[1] - y[0]
        origin = [x.min(), y.min(), 0]
        spacing = [dx, dy, 1]
        ds = src.dataset
        self.assertEqual(np.all(ds.origin == origin), True)
        self.assertEqual(np.allclose(ds.spacing, spacing), True)

        sc = src.dataset.point_data.scalars.to_array()
        s1 = s.transpose()
        self.assertEqual(np.alltrue(sc.ravel() == s1.ravel()), True)

    def test_reset(self):
        "Test the reset method."
        x, y, s, src = self.get_data()

        self.check_traits()
        self.check_dataset()

        # Call reset again with just a few things changed to see if it
        # works correctly.

        s *= 10
        src.reset(x=x, y=y, scalars=s)

        self.check_traits()
        self.check_dataset()

    def test_handlers(self):
        "Test if the various static handlers work correctly."
        x, y, s, src = self.get_data()
        s *= 2
        src.scalars = s

        self.check_traits()
        self.check_dataset()

    def test_set(self):
        "Test if the set method works correctly."
        x, y, s, src = self.get_data()
        s *= 2
        src.trait_set(x=x, y=y, scalars=s)

        self.check_traits()
        self.check_dataset()


class TestMTriangularMeshSource(unittest.TestCase):
    def setUp(self):
        x, y, z = np.array([0, 0, 0]), np.array([0, 0, 1]), np.array([0, 1, 1])
        s = np.array([0.1, 0.2, 0.3])
        self.x, self.y, self.z, self.s = x, y, z, s
        self.triangles = triangles = np.array([[0, 1, 2]])

        src = sources.MTriangularMeshSource()
        src.reset(x=x, y=y, z=z, triangles=triangles, scalars=s)
        self.src = src

    def tearDown(self):
        return

    def get_data(self):
        return self.x, self.y, self.z, self.triangles, self.s, self.src

    def check_traits(self):
        """Check if the sources traits are set correctly."""
        x, y, z, triangles, s, src = self.get_data()

        # Check if points are set correctly.
        self.assertEqual(np.alltrue(src.x == x), True)
        self.assertEqual(np.alltrue(src.y == y), True)
        self.assertEqual(np.alltrue(src.z == z), True)
        # Check the scalars.
        self.assertEqual(np.alltrue(src.scalars == s), True)

    def test_reset(self):
        "Test the reset method."

        x, y, z, triangles, s, src = self.get_data()
        self.check_traits()

        # Call reset again with just a few things changed to see if it
        # works correctly.
        x *= 5
        s *= 10
        src.reset(x=x, y=y, z=z, triangles=triangles, scalars=s)

        self.check_traits()

    def test_changed_size(self):
        """ Change the number of the points, and establish
            to new points, to check that we don't get errors with the
            dimensions of the scalars.
        """
        n = 100
        _, _, _, _, _, src = self.get_data()
        triangles = np.c_[np.arange(n - 3),
                          np.arange(n - 3) + 1, n - 1 - np.arange(n - 3)]
        x, y, z = np.random.random((3, n))
        src.reset(x=x, y=y, z=z, triangles=triangles)

    def test_handlers(self):
        "Test if the various static handlers work correctly."
        x, y, z, triangles, s, src = self.get_data()
        x *= 2
        y *= 2
        s *= 2
        src.x = x
        src.y = y
        src.scalars = s
        src.triangles = triangles

        self.check_traits()

    def test_set(self):
        "Test if the set method works correctly."
        x, y, z, triangles, s, src = self.get_data()
        x *= 2
        s *= 2
        src.trait_set(x=x, scalars=s)

        self.check_traits()

        y *= 9
        s *= 2
        src.trait_set(y=y, scalars=s)

        self.check_traits()

    @patch('mayavi.tools.engine_manager.options.backend', 'test')
    def test_reset_changes_pipeline(self):
        # Given
        from mayavi import mlab
        obj = mlab.triangular_mesh(
            self.x, self.y, self.z, self.triangles, scalars=self.s
        )
        bounds = obj.actor.actor.bounds

        # When
        n = 10
        x, y, z = np.random.random((3, n))
        triangles = np.c_[np.arange(n-3),
                          np.arange(n-3)+1,
                          n-1-np.arange(n-3)]

        obj.mlab_source.reset(x=x, y=y, z=z, triangles=triangles, scalars=z)

        # Then
        self.assertFalse(np.allclose(bounds, obj.actor.actor.bounds))


if __name__ == '__main__':
    unittest.main()



"""
Metadata for all filters.
"""
# Author: Prabhu Ramachandran <prabhu@aero.iitb.ac.in>
# Copyright (c) 2008, Prabhu Ramachandran Enthought, Inc.
# License: BSD Style.

# Local imports.
from mayavi.core.metadata import FilterMetadata
from mayavi.core.pipeline_info import PipelineInfo

BASE = 'mayavi.filters'

################################################################################
# Factory functions.
################################################################################
def make_user_defined_filter():
    from mayavi.filters.user_defined import UserDefined
    f = UserDefined()
    f.setup_filter()
    return f


################################################################################
# Metadata.

cell_derivatives_filter = FilterMetadata(
    id            = "CellDerivativesFilter",
    menu_name          = "&CellDerivatives",
    class_name = BASE + '.cell_derivatives.CellDerivatives',
    tooltip = "Calculate derivatives of input point/vector data and output these as cell data",
    desc = "Calculate derivatives of input point/vector data and output these as cell data",
    help = "Calculate derivatives of input point/vector data and output these as cell data",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['any'],
                               attribute_types=['any'],
                               attributes=['any'])
)

cell_to_point_data_filter = FilterMetadata(
    id            = "CellToPointDataFilter",
    menu_name          = "&CellToPointData",
    class_name = BASE + '.cell_to_point_data.CellToPointData',
    tooltip = "Convert cell data to point data for the active data",
    desc = "Convert cell data to point data for the active data",
    help = "Convert cell data to point data for the active data",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['cell'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['any'],
                               attribute_types=['any'],
                               attributes=['any'])
)

clip_filter = FilterMetadata(
    id            = "DataSetClipperFilter",
    menu_name          = "&DataSet Clipper",
    class_name = BASE + '.data_set_clipper.DataSetClipper',
    tooltip = "Clip the input dataset",
    desc = "Clip the input dataset",
    help = "Clip the input dataset",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['unstructured_grid'],
                               attribute_types=['any'],
                               attributes=['any'])
)


contour_filter = FilterMetadata(
    id            = "ContourFilter",
    menu_name          = "&Contour",
    class_name = BASE + '.contour.Contour',
    tooltip = "Compute contours of the input dataset",
    desc = "Compute contours of the input dataset",
    help = "Compute contours of the input dataset",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['point'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['any'],
                               attribute_types=['any'],
                               attributes=['any'])
)


cut_plane_filter = FilterMetadata(
    id            = "CutPlaneFilter",
    menu_name          = "&CutPlane",
    class_name = BASE + '.cut_plane.CutPlane',
    tooltip = "Slice the input dataset with a cut plane",
    desc = "Slice the input dataset with a cut plane",
    help = "Slice the input dataset with a cut plane",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['poly_data'],
                               attribute_types=['any'],
                               attributes=['any'])
)

decimatepro_filter = FilterMetadata(
    id            = "DecimateProFilter",
    menu_name          = "&DecimatePro",
    class_name = BASE + '.decimatepro.DecimatePro',
    tooltip = "Simpilies a mesh using the DecimatePro filter",
    desc = "Simpilies a mesh using the DecimatePro filter",
    help = "Simpilies a mesh using the DecimatePro filter",
    input_info = PipelineInfo(datasets=['poly_data'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['poly_data'],
                               attribute_types=['any'],
                               attributes=['any'])
)

delaunay2d_filter = FilterMetadata(
    id            = "Delaunay2DFilter",
    menu_name          = "&Delaunay2D",
    class_name = BASE + '.delaunay2d.Delaunay2D',
    tooltip = "Perform a 2D Delaunay triangulation for the given data",
    desc = "Perform a 2D Delaunay triangulation for the given data",
    help = "Perform a 2D Delaunay triangulation for the given data",
    input_info = PipelineInfo(datasets=['structured_grid', 'poly_data',
                                        'unstructured_grid'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['poly_data'],
                               attribute_types=['any'],
                               attributes=['any'])
)

delaunay3d_filter = FilterMetadata(
    id            = "Delaunay3DFilter",
    menu_name          = "Delaunay&3D",
    class_name = BASE + '.delaunay3d.Delaunay3D',
    tooltip = "Perform a 3D Delaunay triangulation for the given data",
    desc = "Perform a 3D Delaunay triangulation for the given data",
    help = "Perform a 3D Delaunay triangulation for the given data",
    input_info = PipelineInfo(datasets=['structured_grid', 'poly_data',
                                        'unstructured_grid'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['unstructured_grid'],
                               attribute_types=['any'],
                               attributes=['any'])
)

elevation_filter = FilterMetadata(
    id            = "ElevationFilter",
    menu_name          = "Elevation Filter",
    class_name = BASE + '.elevation_filter.ElevationFilter',
    tooltip = "Creates scalar data from the elevation along a direction",
    desc = "Creates scalar data from the elevation along a direction",
    help = "Creates scalar data from the elevation along a direction",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['any'],
                               attribute_types=['any'],
                               attributes=['any'])
)

extract_edges_filter = FilterMetadata(
    id            = "ExtractEdgesFilter",
    menu_name          = "Extract Edges",
    class_name = BASE + '.extract_edges.ExtractEdges',
    tooltip = "Turns edges into lines.",
    desc = "Turns edges into lines.",
    help = "Turns edges into lines.",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['poly_data'],
                               attribute_types=['any'],
                               attributes=['any'])
)

extract_grid_filter = FilterMetadata(
    id            = "ExtractGridFilter",
    menu_name          = "Extract &Grid",
    class_name = BASE + '.extract_grid.ExtractGrid',
    tooltip = "Extract/subsample part of any structured grid",
    desc = "Extract/subsample part of any structured grid",
    help = "Extract/subsample part of any structured grid",
    input_info = PipelineInfo(datasets=['image_data',
                                        'rectilinear_grid',
                                        'structured_grid'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['image_data',
                                         'rectilinear_grid',
                                         'structured_grid'],
                               attribute_types=['any'],
                               attributes=['any'])
)

extract_tensor_components_filter = FilterMetadata(
    id            = "ExtractTensorComponentsFilter",
    menu_name          = "Extract &Tensor Components",
    class_name = BASE + '.extract_tensor_components.ExtractTensorComponents',
    tooltip = "Extract tensor components from tensor data",
    desc = "Extract tensor components from tensor data",
    help = "Extract tensor components from tensor data",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['tensors']),
    output_info = PipelineInfo(datasets=['any'],
                               attribute_types=['any'],
                               attributes=['any'])
)

extract_unstructured_grid_filter = FilterMetadata(
    id            = "ExtractUnstructuredGridFilter",
    menu_name          = "Extract &Unstructured Grid",
    class_name = BASE + '.extract_unstructured_grid.ExtractUnstructuredGrid',
    tooltip = "Extract part of an unstructured grid",
    desc = "Extract part of an unstructured grid",
    help = "Extract part of an unstructured grid",
    input_info = PipelineInfo(datasets=['unstructured_grid'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['unstructured_grid'],
                               attribute_types=['any'],
                               attributes=['any'])
)

extract_vector_norm_filter = FilterMetadata(
    id            = "ExtractVectorNormFilter",
    menu_name          = "Extract Vector &Norm",
    class_name = BASE + '.extract_vector_norm.ExtractVectorNorm',
    tooltip = "Compute the vector norm for the current vector data",
    desc = "Compute the vector norm for the current vector data",
    help = "Compute the vector norm for the current vector data",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['vectors']),
    output_info = PipelineInfo(datasets=['any'],
                               attribute_types=['any'],
                               attributes=['any'])
)

extract_vector_components_filter = FilterMetadata(
    id            = "ExtractVectorComponentsFilter",
    menu_name          = "Extract &Vector Components",
    class_name = BASE + '.extract_vector_components.ExtractVectorComponents',
    tooltip = "Extract vector components from vector data",
    desc = "Extract vector components from vector data",
    help = "Extract vector components from vector data",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['vectors']),
    output_info = PipelineInfo(datasets=['any'],
                               attribute_types=['any'],
                               attributes=['any'])
)

gaussian_splatter_filter = FilterMetadata(
    id            = "GaussianSplatterFilter",
    menu_name          = "Gaussian Splatter",
    class_name = BASE + '.gaussian_splatter.GaussianSplatter',
    tooltip = "Builds a structured set of points from a cloud of points, the local density defining the scalar",
    desc = "Builds a structured set of points from a cloud of points, the local density defining the scalar",
    help = """Builds a structured set of points from a cloud of points,
    the local density defining the scalar.  It is essentially equivalent
    to a 3D Gaussian kernel density estimate.""",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['image_data'],
                               attribute_types=['any'],
                               attributes=['any'])
)

greedy_terrain_decimation_filter = FilterMetadata(
    id            = "GreedyTerrainDecimationFilter",
    menu_name          = "Greedy Terrain Decimation",
    class_name = BASE + '.greedy_terrain_decimation.GreedyTerrainDecimation',
    tooltip = "Simplifies image data and performs a triangulation",
    desc = "Simplifies image data and performs a triangulation",
    help = "Simplifies image data and performs a triangulation",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['poly_data'],
                               attribute_types=['any'],
                               attributes=['any'])
)

image_change_information_filter = FilterMetadata(
    id            = "ImageChangeInformationFilter",
    menu_name          = "Change &ImageData information",
    class_name = BASE + '.image_change_information.ImageChangeInformation',
    tooltip = "Change the origin, spacing and extents of an image dataset",
    desc = "Change the origin, spacing and extents of an image dataset",
    help = "Change the origin, spacing and extents of an image dataset",
    input_info = PipelineInfo(datasets=['image_data'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['image_data'],
                               attribute_types=['any'],
                               attributes=['any'])
)

image_data_probe_filter = FilterMetadata(
    id            = "ImageDataProbeFilter",
    menu_name          = "&Probe data onto image data",
    class_name = BASE + '.image_data_probe.ImageDataProbe',
    tooltip = "Samples arbitrary datasets onto an image dataset (cube of data)",
    desc = "Samples arbitrary datasets onto an image dataset (cube of data)",
    help = "Samples arbitrary datasets onto an image dataset (cube of data)",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['image_data'],
                               attribute_types=['any'],
                               attributes=['any'])
)

mask_points_filter = FilterMetadata(
    id            = "MaskPointsFilter",
    menu_name          = "&Mask Points",
    class_name = BASE + '.mask_points.MaskPoints',
    tooltip = "Mask the input points in the data",
    desc = "Mask the input points in the data",
    help = "Mask the input points in the data",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['poly_data'],
                               attribute_types=['any'],
                               attributes=['any'])
)

point_to_cell_data_filter = FilterMetadata(
    id            = "PointToCellDataFilter",
    menu_name          = "&PointToCellData",
    class_name = BASE + '.point_to_cell_data.PointToCellData',
    tooltip = "Convert point data to cell data for the active data",
    desc = "Convert point data to cell data for the active data",
    help = "Convert point data to cell data for the active data",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['point'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['any'],
                               attribute_types=['cell'],
                               attributes=['any'])
)

poly_data_normals_filter = FilterMetadata(
    id            = "PolyDataNormalsFilter",
    menu_name          = "Compute &Normals",
    class_name = BASE + '.poly_data_normals.PolyDataNormals',
    tooltip = "Compute normals and smooth the appearance",
    desc = "Compute normals and smooth the appearance",
    help = "Compute normals and smooth the appearance",
    input_info = PipelineInfo(datasets=['poly_data'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['poly_data'],
                               attribute_types=['any'],
                               attributes=['any'])
)

quadric_decimation_filter = FilterMetadata(
    id            = "QuadricDecimationFilter",
    menu_name          = "Quadric Decimation",
    class_name = BASE + '.quadric_decimation.QuadricDecimation',
    tooltip = "Simplifies a triangular mesh",
    desc = "Simplifies a triangular mesh",
    help = "Simplifies a triangular mesh",
    input_info = PipelineInfo(datasets=['poly_data'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['poly_data'],
                               attribute_types=['any'],
                               attributes=['any'])
)

select_output_filter = FilterMetadata(
    id            = "SelectOutputFilter",
    menu_name          = "&Select Output",
    class_name = BASE + '.select_output.SelectOutput',
    tooltip = "Choose the output of the source that should be used",
    desc = "Choose the output of the source that should be used",
    help = "Choose the output of the source that should be used",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['any'],
                               attribute_types=['any'],
                               attributes=['any'])
)

set_active_attribute_filter = FilterMetadata(
    id            = "SetActiveAttributeFilter",
    menu_name          = "&SetActiveAttribute",
    class_name = BASE + '.set_active_attribute.SetActiveAttribute',
    tooltip = "Set the active attribute (scalar/vector/tensor) to use",
    desc = "Set the active attribute (scalar/vector/tensor) to use",
    help = "Set the active attribute (scalar/vector/tensor) to use",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['any'],
                               attribute_types=['any'],
                               attributes=['any'])
)

stripper = FilterMetadata(
    id            = "Stripper",
    menu_name          = "Stripper",
    class_name = BASE + '.stripper.Stripper',
    tooltip = "Regularizes surfaces by creating triangle strips",
    desc =    "Regularizes surfaces by creating triangle strips",
    help =    "Regularizes surfaces by creating triangle strips",
    input_info = PipelineInfo(datasets=['poly_data'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['poly_data'],
                               attribute_types=['any'],
                               attributes=['any'])
)

transform_data_filter = FilterMetadata(
    id            = "TransformDataFilter",
    menu_name          = "T&ransform Data",
    class_name = BASE + '.transform_data.TransformData',
    tooltip = "Transform (rotate/translate/scale) non ImageData datasets",
    desc = "Transform (rotate/translate/scale) non ImageData datasets",
    help = "Transform (rotate/translate/scale) non ImageData datasets",
    input_info = PipelineInfo(datasets=['poly_data',
                                        'structured_grid',
                                        'unstructured_grid'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['poly_data',
                                         'structured_grid',
                                         'unstructured_grid'],
                               attribute_types=['any'],
                               attributes=['any'])
)

threshold_filter = FilterMetadata(
    id            = "ThresholdFilter",
    menu_name          = "&Threshold",
    class_name = BASE + '.threshold.Threshold',
    tooltip = "Threshold input data based on scalar values",
    desc = "Threshold input data based on scalar values",
    help = "Threshold input data based on scalar values",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['poly_data',
                                         'unstructured_grid'],
                               attribute_types=['any'],
                               attributes=['any'])
)

triangle_filter = FilterMetadata(
    id            = "TriangleFilterFilter",
    menu_name          = "TriangleFilter",
    class_name = BASE + '.triangle_filter.TriangleFilter',
    tooltip = "Convert input polygons and triangle strips to triangles",
    desc = "Convert input polygons and triangle strips to triangles",
    help = "Convert input polygons and triangle strips to triangles",
    input_info = PipelineInfo(datasets=['poly_data'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['poly_data'],
                               attribute_types=['any'],
                               attributes=['any'])
)

tube_filter = FilterMetadata(
    id            = "TubeFilter",
    menu_name          = "Tu&be",
    class_name = BASE + '.tube.Tube',
    tooltip = "Turns lines into tubes",
    desc = "Turns lines into tubes",
    help = "Turns lines into tubes",
    input_info = PipelineInfo(datasets=['poly_data'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['poly_data'],
                               attribute_types=['any'],
                               attributes=['any'])
)

user_defined_filter = FilterMetadata(
    id            = "UserDefinedFilter",
    menu_name          = "&UserDefined",
    factory = make_user_defined_filter,
    tooltip = "Create a UserDefined filter (will popup a selection dialog)",
    desc = "Create a UserDefined filter (will popup a selection dialog)",
    help = "Create a UserDefined filter (will popup a selection dialog)",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['any']),
    output_info = PipelineInfo(datasets=['any'],
                               attribute_types=['any'],
                               attributes=['any'])
)

vorticity_filter = FilterMetadata(
    id            = "VorticityFilter",
    menu_name          = "&Vorticity",
    class_name = BASE + '.vorticity.Vorticity',
    tooltip = "Calculate the vorticity (curl) of input vector field",
    desc = "Calculate the vorticity (curl) of input vector field",
    help = "Calculate the vorticity (curl) of input vector field",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['vectors']),
    output_info = PipelineInfo(datasets=['any'],
                               attribute_types=['any'],
                               attributes=['any'])
)

warp_scalar_filter = FilterMetadata(
    id            = "WarpScalarFilter",
    menu_name          = "Warp S&calar",
    class_name = BASE + '.warp_scalar.WarpScalar',
    tooltip = "Move points of data along normals by the scalar data",
    desc = "Move points of data along normals by the scalar data",
    help = "Move points of data along normals by the scalar data",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['scalars']),
    output_info = PipelineInfo(datasets=['any'],
                               attribute_types=['any'],
                               attributes=['any'])
)

warp_vector_filter = FilterMetadata(
    id            = "WarpVectorFilter",
    menu_name          = "Warp &Vector",
    class_name = BASE + '.warp_vector.WarpVector',
    tooltip = "Move points of data along the vector data at point",
    desc = "Move points of data along the vector data at point",
    help = "Move points of data along the vector data at point",
    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['vectors']),
    output_info = PipelineInfo(datasets=['any'],
                               attribute_types=['any'],
                               attributes=['any'])
)

# Now collect all the filters for the mayavi registry.
filters = [cell_derivatives_filter,
           cell_to_point_data_filter,
           clip_filter,
           contour_filter,
           cut_plane_filter,
           decimatepro_filter,
           delaunay2d_filter,
           delaunay3d_filter,
           elevation_filter,
           extract_edges_filter,
           extract_grid_filter,
           extract_tensor_components_filter,
           extract_unstructured_grid_filter,
           extract_vector_norm_filter,
           extract_vector_components_filter,
           gaussian_splatter_filter,
           greedy_terrain_decimation_filter,
           image_change_information_filter,
           image_data_probe_filter,
           mask_points_filter,
           point_to_cell_data_filter,
           poly_data_normals_filter,
           quadric_decimation_filter,
           select_output_filter,
           set_active_attribute_filter,
           stripper,
           transform_data_filter,
           threshold_filter,
           triangle_filter,
           tube_filter,
           user_defined_filter,
           vorticity_filter,
           warp_scalar_filter,
           warp_vector_filter,
        ]



"""An Image file reader object.

"""
# Author:  KK Rai (kk.rai [at] iitb.ac.in)
#          R. Ambareesha (ambareesha [at] iitb.ac.in)
#          Chandrashekhar Kaushik
#          Suyog Dutt Jain <suyog.jain [at] aero.iitb.ac.in>
#          Prabhu Ramachandran <prabhu [at] aero.iitb.ac.in>
# Copyright (c) 2007-2020, Enthought, Inc.
# License: BSD Style.

from os.path import basename

# Enthought library imports.
from traits.api import Instance, Str, Dict
from traitsui.api import View, Group, Item, Include
from tvtk.api import tvtk

# Local imports.
from mayavi.core.file_data_source import FileDataSource
from mayavi.core.pipeline_info import PipelineInfo


########################################################################
# `ImageReader` class
########################################################################
class ImageReader(FileDataSource):

    """A Image file reader. The reader supports all the
    different types of Image files.
    """

    # The version of this class.  Used for persistence.
    __version__ = 0

    # The Image data file reader.
    reader = Instance(tvtk.Object, allow_none=False, record=True)

    # Information about what this object can produce.
    output_info = PipelineInfo(datasets=['image_data'])

    # Our view.
    view = View(Group(Include('time_step_group'),
                      Item(name='base_file_name'),
                      Item(name='reader',
                           style='custom',
                           resizable=True),
                      show_labels=False),
                resizable=True)

    ######################################################################
    # Private Traits
    _image_reader_dict = Dict(Str, Instance(tvtk.Object))

    ######################################################################
    # `object` interface
    ######################################################################
    def __init__(self, **traits):
        d = {'bmp':tvtk.BMPReader(),
             'jpg':tvtk.JPEGReader(),
             'png':tvtk.PNGReader(),
             'pnm':tvtk.PNMReader(),
             'dcm':tvtk.DICOMImageReader(),
             'tiff':tvtk.TIFFReader(),
             'ximg':tvtk.GESignaReader(),
             'dem':tvtk.DEMReader(),
             'mha':tvtk.MetaImageReader(),
             'mhd':tvtk.MetaImageReader(),
            }
        # Account for pre 5.2 VTk versions, without MINC reader
        if hasattr(tvtk, 'MINCImageReader'):
            d['mnc'] = tvtk.MINCImageReader()
        d['jpeg'] = d['jpg']
        self._image_reader_dict = d
        # Call parent class' init.
        super(ImageReader, self).__init__(**traits)

    def __set_pure_state__(self, state):
        # The reader has its own file_name which needs to be fixed.
        state.reader.file_name = state.file_path.abs_pth
        # Now call the parent class to setup everything.
        super(ImageReader, self).__set_pure_state__(state)

    ######################################################################
    # `FileDataSource` interface
    ######################################################################
    def update(self):
        self.reader.update()
        if len(self.file_path.get()) == 0:
            return
        self.render()

    def has_output_port(self):
        """ Return True as the reader has output port."""
        return True

    def get_output_object(self):
        """ Return the reader output port."""
        return self.reader.output_port

    ######################################################################
    # Non-public interface
    ######################################################################
    def _file_path_changed(self, fpath):
        value = fpath.get()
        if len(value) == 0:
            return
        # Extract the file extension
        splitname = value.strip().split('.')
        extension = splitname[-1].lower()
        # Select image reader based on file type
        old_reader = self.reader
        if extension in self._image_reader_dict:
            self.reader = self._image_reader_dict[extension]
        else:
            self.reader = tvtk.ImageReader()

        self.reader.file_name = value.strip()
        self.reader.update()
        self.reader.update_information()

        if old_reader is not None:
            old_reader.on_trait_change(self.render, remove=True)
        self.reader.on_trait_change(self.render)

        self.outputs = [self.reader]

        # Change our name on the tree view
        self.name = self._get_name()

    def _get_name(self):
        """ Returns the name to display on the tree view.  Note that
        this is not a property getter.
        """
        fname = basename(self.file_path.get())
        ret = "%s"%fname
        if len(self.file_list) > 1:
            ret += " (timeseries)"
        if '[Hidden]' in self.name:
            ret += ' [Hidden]'

        return ret



"""
Base class for factories for adding objects to the pipeline.

"""

# Author: Gael Varoquaux <gael.varoquaux@normalesup.org>
# Copyright (c) 2007-2020, Enthought, Inc.
# License: BSD Style.

import warnings

from traits.api import HasPrivateTraits, Str, TraitError,\
            Instance, Any, Bool
from mayavi.core.filter import Filter
from mayavi.core.engine import Engine
from mayavi.core.source import Source
from mayavi.core.scene import Scene
from mayavi.core.module_manager import ModuleManager

from tvtk.api import tvtk

from .auto_doc import make_doc
from . import tools
from .engine_manager import get_engine


def get_obj(obj, components):
    """ Get the target object for the specified components. """

    for component in components:
        obj = getattr(obj, component)

    return obj


def make_function(factory_class):
    def the_function(*args, **kwargs):
        factory = factory_class(*args, **kwargs)
        return factory._target

    the_function.__doc__ = make_doc(factory_class)
    the_function.__name__ = factory_class.__name__.lower()
    return the_function


def get_module_manager(obj):
    """ Returns the module manager that would be used when a module
        is added on the given object, if any, and None elsewhere.
    """
    if hasattr(obj, 'module_manager'):
        return obj.module_manager
    elif isinstance(obj, ModuleManager):
        return obj
    for child in reversed(obj.children):
        if isinstance(child, ModuleManager):
            return child
    else:
        return None


##############################################################################
class PipeFactory(HasPrivateTraits):
    """ Base class for all factories adding pipes on the pipeline """

    name = Str(adapts='name', help='the name of the vtk object created.')

    figure = Instance(Scene)

    _engine = Instance(Engine, help=('the figure on which the object '
                'should be added'))

    _target = Any

    _do_redraw = Bool

    def add_module(self, parent, kwargs=dict()):
        """ Add the target module to the given object.
        """
        # We check to see if the module-manager-related option require to
        # add a new module manager:
        if parent is not None:
            module_manager = get_module_manager(parent)
            if (module_manager is not None and
                        len(module_manager.children) > 0):
                scalar_lut = module_manager.scalar_lut_manager
                vector_lut = module_manager.vector_lut_manager
                if 'vmin' in kwargs:
                    if not scalar_lut.use_default_range and \
                            kwargs['vmin'] != scalar_lut.data_range[0]:
                        parent = self._engine.add_module(ModuleManager(),
                                                        module_manager.parent)
                    elif not scalar_lut.use_default_range and \
                            kwargs['vmin'] != scalar_lut.data_range[0]:
                        parent = self._engine.add_module(ModuleManager(),
                                                        module_manager.parent)

                elif 'vmax' in kwargs:
                    if not scalar_lut.use_default_range and \
                            kwargs['vmax'] != scalar_lut.data_range[1]:
                        parent = self._engine.add_module(ModuleManager(),
                                                        module_manager.parent)
                    elif not scalar_lut.use_default_range and \
                            kwargs['vmax'] != scalar_lut.data_range[1]:
                        parent = self._engine.add_module(ModuleManager(),
                                                        module_manager.parent)

                elif 'colormap' in kwargs:
                    cmap = kwargs['colormap']
                    if (scalar_lut.lut_mode != cmap
                                        or vector_lut.lut_mode != cmap):
                        parent = self._engine.add_module(ModuleManager(),
                                            module_manager.parent)

        self._engine.add_module(self._target, obj=parent)

    def __init__(self, parent, **kwargs):
        # We are not passing the traits to the parent class
        super(PipeFactory, self).__init__()
        # Try to find the right engine and scene to work with
        ancester = parent
        while hasattr(ancester, 'parent'):
            ancester = getattr(ancester, 'parent')
            if isinstance(ancester, Scene):
                self._scene = ancester
                self._engine = ancester.parent
                break
        else:
            if self.figure is not None:
                self._scene = self.figure
            else:
                self._scene = tools.gcf()
                self._engine = get_engine()
        scene = self._scene.scene
        if self.figure is not None and self.figure is not self._scene:
            warnings.warn('Trying to add a module on the wrong scene')
        if isinstance(parent, (Source, tvtk.DataSet)) \
                and not isinstance(parent, Filter) and self._scene is not None:
            # Search the current scene to see if the  source is already
            # in it, if not add it.
            if not parent in self._scene.children:
                parent = tools.add_dataset(parent, figure=self._scene)

        if scene is not None:
            self._do_redraw = not scene.disable_render
            scene.disable_render = True
        if issubclass(self._target.__class__, Filter):
            self._engine.add_filter(self._target, obj=parent)
        else:
            self.add_module(parent, kwargs)

        # Inject the magical mlab source trait.
        if hasattr(parent, 'mlab_source'):
            ms = parent.mlab_source
            self._target.add_trait('mlab_source', Instance(ms.__class__))
            self._target.mlab_source = ms

        traits = self.trait_get(self.class_trait_names())
        [traits.pop(key) for key in list(traits.keys())
                                    if key[0] == '_' or key is None]
        traits.update(kwargs)
        # Now calling the traits setter, so that traits handlers are
        # called
        self.trait_set(**traits)
        if scene is not None:
            scene.disable_render = not self._do_redraw

    def set(self, trait_change_notify=True, **traits):
        return self.trait_set(trait_change_notify=trait_change_notify,
                              **traits)

    def trait_set(self, trait_change_notify=True, **traits):
        """ Same as HasTraits.set except that notification is forced,
        unless trait_change_notify==False"""
        HasPrivateTraits.trait_set(
            self, trait_change_notify=trait_change_notify, **traits)
        if trait_change_notify == False:
            return
        for trait in traits:
            callback = getattr(self, '_%s_changed' % trait)
            value = getattr(self, trait)
            try:
                if callback is not None:
                    callback()
                self._anytrait_changed(trait, value)
            except TraitError:
                if value is None:
                    # This means "default"
                    pass
                else:
                    raise

    def _anytrait_changed(self, name, value):
        """ This is where we implement the adaptation code. """
        trait = self.trait(name)
        if name[0] == '_':
            # Private attribute
            return
        # hasattr(traits, "adapts") always returns True :-<.
        if not trait.adapts is None:
            components = trait.adapts.split('.')
            obj = get_obj(self._target, components[:-1])
            setattr(obj, components[-1], value)



"""Elementary tests for wrapper_gen.py.

This test suite is not comprehensive because it is very hard to test
wrapper_gen directly.

"""
# Author: Prabhu Ramachandran <prabhu_r@users.sf.net>
# Copyright (c) 2004, Prabhu Ramachandran,  Enthought, Inc.
# License: BSD Style.

import unittest
import vtk

from tvtk import wrapper_gen

_cache = wrapper_gen.WrapperGenerator()


class TestWrapperGenerator(unittest.TestCase):
    def setUp(self):
        self.wg = _cache

    def test_find_type(self):
        wg = self.wg
        sigs = ['int', 'vtkOpenGLVolumeMapper',
                ('int', 'int', 'float', 'list'),
                ('int', 'vtkActor', 'vtkXMLReader'),
                ['vtkImageActor', 'vtkExporter'],
                ['int', 'vtkDataArray', 'vtkCellArray',
                 'vtkIdTypeArray']
                ]
        expect = ['basic', 'vtk', 'basic', 'vtk', 'vtk', 'array']
        for i, sig in enumerate(sigs):
            self.assertEqual(expect[i], wg._find_type(sig))

    def test_sig_types(self):
        wg = self.wg
        meths = [vtk.vtkProperty.GetColor,
                 vtk.vtkProperty.GetRepresentation,
                 vtk.vtkStructuredPointsReader.GetOutput,
                 vtk.vtkPolyData.SetPoints,
                 vtk.vtkPolyData.SetPolys,
                 vtk.vtkQuad.CellBoundary,
                 vtk.vtkContourFilter.SetLocator
                 ]
        expect = [('basic', 'basic'),
                  ('basic', None),
                  ('vtk', 'basic'),
                  ('basic', 'array'),
                  ('basic', 'array'),
                  ('basic', 'array'),
                  ('basic', 'vtk'),
                  ]
        for i, meth in enumerate(meths):
            sig = wg.parser.get_method_signature(meth)
            self.assertEqual(expect[i], wg._find_sig_type(sig))
        sig = [(['int'], ['int']),
               (['vtkStructuredPoints'], ['vtkFooClass'])]
        self.assertEqual(('vtk', 'vtk'), wg._find_sig_type(sig))

    def test_unicode_return_value(self):
        wg = self.wg
        meth = vtk.vtkDelimitedTextReader.GetUnicodeRecordDelimiters
        expect = 'unicode' if '-> unicode' in meth.__doc__ else 'string'
        sig = wg.parser.get_method_signature(meth)
        self.assertEqual(sig[0][0][0], expect)


if __name__ == "__main__":
    unittest.main()



"""Actions to start various filters.
"""
# Author: Prabhu Ramachandran <prabhu_r@users.sf.net>
# Copyright (c) 2005-2020, Enthought, Inc.
# License: BSD Style.

from pyface.action.api import Action
from traits.api import Instance

from mayavi.plugins.script import get_imayavi
from mayavi.core.registry import registry
from mayavi.core.metadata import Metadata
from mayavi.core.pipeline_base import PipelineBase

def new_class(name, bases, dict_):
    try:
        import new
        return new.classobj(name, bases, dict_)
    except ImportError:
        return type(name, bases, dict_)

######################################################################
# `FilterAction` class.
######################################################################
class FilterAction(Action):

    # The Metadata associated with this particular action.
    metadata = Instance(Metadata)

    mayavi = Instance('mayavi.plugins.script.Script')

    # We disable the actions by default since these are dynamically
    # enabled depending on the current selection or object.
    enabled = False

    def __init__(self, **traits):
        super(FilterAction, self).__init__(**traits)
        self.mayavi.engine.on_trait_change(self._update_enabled,
                                           ['current_selection',
                                            'current_object'])

    ###########################################################################
    # 'Action' interface.
    ###########################################################################
    def perform(self, event):
        """ Performs the action. """
        callable = self.metadata.get_callable()
        obj = callable()
        mv = self.mayavi
        mv.add_filter(obj)
        mv.engine.current_selection = obj

    def _update_enabled(self, obj):
        if isinstance(obj, PipelineBase):
            e = obj.menu_helper.check_active(self.metadata)
            self.enabled = e
        else:
            self.enabled = False

    def _mayavi_default(self):
        return get_imayavi(self.window)


######################################################################
# Creating the filter actions automatically.
for filter in registry.filters:
    d = {'tooltip': filter.tooltip,
         'description': filter.desc,
         'metadata': filter}
    action = new_class(filter.id, (FilterAction,), d)
    globals()[filter.id] = action



# Author: Gael Varoquaux <gael.varoquaux@normalesup.org>
# Copyright (c) 2007, Enthought, Inc.
# License: BSD Style.

# Enthought library imports.
from traits.api import Instance
from tvtk.api import tvtk

# Local imports
from mayavi.filters.filter_base import FilterBase
from mayavi.core.pipeline_info import PipelineInfo


######################################################################
# `GaussianSplatter` class.
######################################################################
class GaussianSplatter(FilterBase):


    # The version of this class.  Used for persistence.
    __version__ = 0

    # The actual TVTK filter that this class manages.
    filter = Instance(tvtk.GaussianSplatter, args=(),
                      allow_none=False, record=True)

    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['any'])

    output_info = PipelineInfo(datasets=['image_data'],
                               attribute_types=['any'],
                               attributes=['any'])




"""A VTK XML file reader object.
"""
# Author: Prabhu Ramachandran <prabhu_r@users.sf.net>
# Copyright (c) 2005, Enthought, Inc.
# License: BSD Style.

# Standard library imports.
from os.path import basename

# Enthought library imports.
from traits.api import Instance, List, Str, Bool, Button
from traitsui.api import View, Group, Item, Include
from tvtk.api import tvtk

# Local imports.
from mayavi.core.common import error
from mayavi.core.file_data_source import FileDataSource
from mayavi.core.trait_defs import DEnum
from mayavi.core.pipeline_info import (PipelineInfo,
        get_tvtk_dataset_name)


######################################################################
# Utility functions.
######################################################################
def find_file_data_type(file_name):
    "Parses the named file to see what type of data there is."
    r = tvtk.XMLFileReadTester(file_name=file_name)
    if r.test_read_file():
        return r.file_data_type
    else:
        error("File %s is not a valid VTK XML file!"%(file_name))


def get_array_type(arr):
    """Returns if the array is a scalar ('scalars'), vector
    ('vectors') or tensor ('tensors').  It looks at the number of
    components to decide.  If it has a wierd number of components it
    returns the empty string.
    """
    n = arr.number_of_components
    ret = {1: 'scalars', 3: 'vectors', 4: 'scalars', 9:'tensors'}
    return ret.get(n) or ''


def get_attribute_list(data):
    """ Gets scalar, vector and tensor information from the given data
    (either cell or point data).
    """
    attr = {'scalars':[], 'vectors':[], 'tensors':[]}
    if data is not None:
        n = data.number_of_arrays
        for i in range(n):
            name = data.get_array_name(i)
            arr = data.get_array(i)
            if arr is not None:
                # Some VTK datasets claim they have n arrays, but
                # actually some of these are None (eg the output of a
                # tvtk.GraphToPolyData())
                t = get_array_type(arr)
                if len(t) > 0 and name is not None:
                    attr[t].extend([name])

    def _mk_first(lst, value):
        """Makes the specified `value` the first item in `lst`."""
        lst.remove(value)
        lst.insert(0, value)

    attr1 = attr.copy()
    for a in attr:
        v = getattr(data, a)
        if v is not None:
            name = v.name
            if name is not None:
                try:
                    _mk_first(attr[a], v.name)
                except ValueError:
                    # Sometimes we have a multi-component scalar.
                    attr1[a].insert(0, name)
    return attr1


def get_all_attributes(obj):
    """Gets the scalar, vector and tensor attributes that are
    available in the given VTK data object.
    """
    point_attr = get_attribute_list(obj.point_data)
    cell_attr = get_attribute_list(obj.cell_data)
    return point_attr, cell_attr


######################################################################
# `VTKXMLFileReader` class
######################################################################
class VTKXMLFileReader(FileDataSource):

    """A VTK XML file reader.  The reader supports all the different
    types of data sets.  This reader also supports a time series.
    Currently, this reader assumes that there is only one output that
    has configurable attributes.
    """

    # The version of this class.  Used for persistence.
    __version__ = 0

    ########################################
    # Dynamic traits: These traits are dynamic and are automatically
    # updated depending on the contents of the file.

    # The active point scalar name.  An empty string indicates that
    # the attribute is "deactivated".  This is useful when you have
    # both point and cell attributes and want to use cell data by
    # default.
    point_scalars_name = DEnum(values_name='_point_scalars_list',
                               desc='scalar point data attribute to use')
    # The active point vector name.
    point_vectors_name = DEnum(values_name='_point_vectors_list',
                               desc='vectors point data attribute to use')
    # The active point tensor name.
    point_tensors_name = DEnum(values_name='_point_tensors_list',
                               desc='tensor point data attribute to use')

    # The active cell scalar name.
    cell_scalars_name = DEnum(values_name='_cell_scalars_list',
                               desc='scalar cell data attribute to use')
    # The active cell vector name.
    cell_vectors_name = DEnum(values_name='_cell_vectors_list',
                               desc='vectors cell data attribute to use')
    # The active cell tensor name.
    cell_tensors_name = DEnum(values_name='_cell_tensors_list',
                               desc='tensor cell data attribute to use')
    ########################################

    # The VTK data file reader.
    reader = Instance(tvtk.XMLReader)

    refresh = Button('Update reader')

    # Information about what this object can produce.
    output_info = PipelineInfo(datasets=['any'],
                               attribute_types=['any'],
                               attributes=['any'])

    # Our view.
    view = View(Group(Include('time_step_group'),
                      Item(name='point_scalars_name'),
                      Item(name='point_vectors_name'),
                      Item(name='point_tensors_name'),
                      Item(name='cell_scalars_name'),
                      Item(name='cell_vectors_name'),
                      Item(name='cell_tensors_name'),
                      Item(name='reader'),
                      Item(name='refresh', show_label=False)
                      ))

    ########################################
    # Private traits.

    # These private traits store the list of available data
    # attributes.  The non-private traits use these lists internally.
    _point_scalars_list = List(Str)
    _point_vectors_list = List(Str)
    _point_tensors_list = List(Str)
    _cell_scalars_list = List(Str)
    _cell_vectors_list = List(Str)
    _cell_tensors_list = List(Str)

    # This filter allows us to change the attributes of the data
    # object and will ensure that the pipeline is properly taken care
    # of.  Directly setting the array in the VTK object will not do
    # this.
    _assign_attribute = Instance(tvtk.AssignAttribute, args=(),
                                 allow_none=False)

    # Toggles if this is the first time this object has been used.
    _first = Bool(True)

    ######################################################################
    # `object` interface
    ######################################################################
    def __get_pure_state__(self):
        d = super(VTKXMLFileReader, self).__get_pure_state__()
        for name in ('_assign_attribute', '_first'):
            d.pop(name, None)
        # Pickle the 'point_scalars_name' etc. since these are
        # properties and not in __dict__.
        attr = {}
        for name in ('point_scalars', 'point_vectors',
                     'point_tensors', 'cell_scalars',
                     'cell_vectors', 'cell_tensors'):
            d.pop('_' + name + '_list', None)
            d.pop('_' + name + '_name', None)
            x = name + '_name'
            attr[x] = getattr(self, x)
        d.update(attr)
        return d

    def __set_pure_state__(self, state):
        # The reader has its own file_name which needs to be fixed.
        state.reader.file_name = state.file_path.abs_pth
        # Now call the parent class to setup everything.
        super(VTKXMLFileReader, self).__set_pure_state__(state)

    ######################################################################
    # `Base` interface
    ######################################################################
    def start(self):
        """This is invoked when this object is added to the mayavi
        pipeline.
        """
        # Do nothing if we are already running.
        if self.running:
            return

        # Call the parent method to do its thing.  This will typically
        # start all our children.
        super(VTKXMLFileReader, self).start()

    def stop(self):
        """Invoked when this object is removed from the mayavi
        pipeline.
        """
        if not self.running:
            return

        # Call the parent method to do its thing.
        super(VTKXMLFileReader, self).stop()


    ######################################################################
    # `FileDataSource` interface
    ######################################################################
    def update(self):
        if len(self.file_path.get()) == 0:
            return
        reader = self.reader
        reader.update()
        self.render()

    def update_data(self):
        if len(self.file_path.get()) == 0:
            return
        self.reader.update()
        pnt_attr, cell_attr = get_all_attributes(self.reader.output)

        def _setup_data_traits(obj, attributes, d_type):
            """Given the object, the dict of the attributes from the
            `get_all_attributes` function and the data type
            (point/cell) data this will setup the object and the data.
            """
            attrs = ['scalars', 'vectors', 'tensors']
            aa = obj._assign_attribute
            data = getattr(obj.reader.output, '%s_data'%d_type)
            for attr in attrs:
                values = attributes[attr]
                values.append('')
                setattr(obj, '_%s_%s_list'%(d_type, attr), values)
                if len(values) > 1:
                    default = getattr(obj, '%s_%s_name'%(d_type, attr))
                    if obj._first and len(default) == 0:
                        default = values[0]
                    getattr(data, 'set_active_%s'%attr)(default)
                    aa.assign(default, attr.upper(),
                              d_type.upper() +'_DATA')
                    aa.update()
                    kw = {'%s_%s_name'%(d_type, attr): default,
                          'trait_change_notify': False}
                    obj.trait_set(**kw)

        _setup_data_traits(self, cell_attr, 'cell')
        _setup_data_traits(self, pnt_attr, 'point')
        if self._first:
            self._first = False
        # Propagate the data changed event.
        self.data_changed = True

    def has_output_port(self):
        """ Return True as the reader has output port."""
        return True

    def get_output_object(self):
        """ Return the reader output port."""
        return self.reader.output_port

    ######################################################################
    # Non-public interface
    ######################################################################
    def _file_path_changed(self, fpath):
        value = fpath.get()
        if len(value) == 0:
            return
        else:
            if self.reader is None:
                d_type = find_file_data_type(fpath.get())
                self.reader = eval('tvtk.XML%sReader()'%d_type)
            reader = self.reader
            reader.file_name = value
            reader.update()
            aa = self._assign_attribute
            self.configure_input(aa, self.reader)
            self.update_data()
            aa.update()
            outputs = [aa]
            self.outputs = outputs

            self.output_info.datasets = [get_tvtk_dataset_name(outputs[0])]

            # Change our name on the tree view
            self.name = self._get_name()

    def _set_data_name(self, data_type, attr_type, value):
        if value is None:
            return

        reader_output = self.reader.output
        if len(value) == 0:
            # If the value is empty then we deactivate that attribute.
            d = getattr(reader_output, attr_type + '_data')
            method = getattr(d, 'set_active_%s'%data_type)
            method(None)
            self.data_changed = True
            return

        aa = self._assign_attribute
        data = None
        if attr_type == 'point':
            data = reader_output.point_data
        elif attr_type == 'cell':
            data = reader_output.cell_data

        method = getattr(data, 'set_active_%s'%data_type)
        method(value)
        aa.assign(value, data_type.upper(), attr_type.upper() +'_DATA')
        aa.update()
        # Fire an event, so the changes propagate.
        self.data_changed = True

    def _point_scalars_name_changed(self, value):
        self._set_data_name('scalars', 'point', value)

    def _point_vectors_name_changed(self, value):
        self._set_data_name('vectors', 'point', value)

    def _point_tensors_name_changed(self, value):
        self._set_data_name('tensors', 'point', value)

    def _cell_scalars_name_changed(self, value):
        self._set_data_name('scalars', 'cell', value)

    def _cell_vectors_name_changed(self, value):
        self._set_data_name('vectors', 'cell', value)

    def _cell_tensors_name_changed(self, value):
        self._set_data_name('tensors', 'cell', value)

    def _get_name(self):
        """ Gets the name to display on the tree view.
        """
        fname = basename(self.file_path.get())
        ret = "VTK XML file (%s)"%fname
        if len(self.file_list) > 1:
            ret += " (timeseries)"
        if '[Hidden]' in self.name:
            ret += ' [Hidden]'

        return ret

    def _refresh_fired(self):
        self.reader.modified()
        self.update_data()



# Standard library imports.
import numpy as np
import unittest
from unittest.mock import patch

# Enthought library imports
from mayavi.tests.common import get_example_data
from mayavi import mlab


class TestVolumeWorksWithProbe(unittest.TestCase):
    def setUp(self):
        self._orig_backend = mlab.options.backend
        mlab.options.backend = "test"

    def tearDown(self):
        mlab.options.backend = self._orig_backend

    def test_volume_works_with_probe(self):
        src = mlab.pipeline.open(get_example_data('pyramid_ug.vtu'))
        idp = mlab.pipeline.image_data_probe(src)
        with patch('pyface.api.error') as m:
            vol = mlab.pipeline.volume(idp)
        self.assertEqual(m.call_count, 0)
        self.assertEqual(
            np.allclose(vol.volume.center, (3.0, 3.0, 1.5)),True
        )



if __name__ == '__main__':
    unittest.main()



# Author: Suyog Dutt Jain <suyog.jain@aero.iitb.ac.in>
# Copyright (c) 2009,  Enthought, Inc.
# License: BSD Style.

# Standard library imports.
import unittest

# Enthought library imports
from mayavi.sources.unstructured_grid_reader import UnstructuredGridReader
from mayavi.tests.data_reader_test_base import DataReaderTestBase

# Local imports.
from mayavi.tests.common import get_example_data


class TestAVSUCDReader(DataReaderTestBase):

    def setup_reader(self):

        """"Setup the reader in here.  This is called after the engine
        has been created and started.  The engine is available as
        self.e.  This method is called by setUp().
        """
        # Read a AVSUCD data file.
        r = UnstructuredGridReader()
        r.initialize(get_example_data('cellsnd.ascii.inp'))
        self.e.add_source(r)
        self.bounds =(-2.0, 2.0, -2.0, 2.0, 0.0, 0.0)

    def test_avsucd_data_reader(self):
        "Test if the test fixture works"
        #Now test.

        self.check(self.scene, self.bounds)

    def test_save_and_restore(self):
        """Test if saving a visualization and restoring it works."""

        self.check_saving(self.e, self.scene, self.bounds)

    def test_deepcopied(self):
        """Test if the MayaVi2 visualization can be deep-copied."""
        ############################################################
        # Test if the MayaVi2 visualization can be deep-copied.

        self.check_deepcopying(self.scene, self.bounds)


#TODO: Update the ExodusIIReader test for scenarios as for the other readers
#in this module.
@unittest.skip("ExodusIIReader support is disabled for now.")
class TestExodusIIReader(DataReaderTestBase):

    def setup_reader(self):

        """"Setup the reader in here.  This is called after the engine
        has been created and started.  The engine is available as
        self.e.  This method is called by setUp().
        """
        # Read a Exodus data file.
        r = UnstructuredGridReader()
        r.initialize(get_example_data('disk_out_ref.ex2'))
        self.e.add_source(r)

    def test_point_cell_data(self):
        """ Test if the point and cell data is parsed correctly."""
        u_grid = self.scene.children[0].outputs[0]
        self.assertEqual(u_grid.point_data.number_of_arrays, 7)
        self.assertEqual(u_grid.cell_data.number_of_arrays, 1)
        self.assertEqual(u_grid.point_data.number_of_tuples, 8499)
        self.assertEqual(u_grid.cell_data.number_of_tuples, 7472)

class TestGambitReader(DataReaderTestBase):

    def setup_reader(self):

        """"Setup the reader in here.  This is called after the engine
        has been created and started.  The engine is available as
        self.e.  This method is called by setUp().
        """
        # Read a Gambit data file.
        r = UnstructuredGridReader()
        r.initialize(get_example_data('prism.neu'))
        self.e.add_source(r)
        self.bounds = (-1.0, 1.0, -1.0, 1.0, 0.0, 1.0)

    def test_gambit_data_reader(self):
        "Test if the test fixture works"
        #Now test.

        self.check(self.scene, self.bounds)

    def test_save_and_restore(self):
        """Test if saving a visualization and restoring it works."""

        self.check_saving(self.e, self.scene, self.bounds)

    def test_deepcopied(self):
        """Test if the MayaVi2 visualization can be deep-copied."""
        ############################################################
        # Test if the MayaVi2 visualization can be deep-copied.

        self.check_deepcopying(self.scene, self.bounds)

if __name__ == '__main__':
    unittest.main()



"""
 A registry for engines, sources, filters and modules.
"""
# Author: Prabhu Ramachandran <prabhu@aero.iitb.ac.in>
# Copyright (c) 2008-2020, Enthought, Inc.
# License: BSD Style.

# Standard library imports.
from os.path import splitext
import logging

# Enthought library imports.
from traits.api import HasTraits, List, Instance, Dict, Str

# Local imports.
from mayavi.core.metadata import Metadata, import_symbol

# A logger for this module.
logger = logging.getLogger(__name__)


################################################################################
# `Registry` class.
################################################################################
class Registry(HasTraits):

    """
    This class is a registry for various engines, and metadata from
    sources, filters and modules
    """

    # The mayavi engines used.
    engines = Dict(Str, Instance('mayavi.core.engine.Engine'))

    # The metadata for the sources.
    sources = List(Metadata)

    # The metadata for the modules.
    modules = List(Metadata)

    # The metadata for the filters.
    filters = List(Metadata)

    ######################################################################
    # `Registry` interface.
    ######################################################################
    def register_engine(self, engine, name=''):

        """Registers a mayavi engine with an optional name.   Note that
        we allow registering an engine with the same name as another
        already registered.  """

        engines = self.engines
        if len(name) == 0:
            name = '%s%d'%(engine.__class__.__name__,
                           len(engines) + 1)

        logger.debug('Engine [%s] named %s registered', engine, name)
        engines[name] = engine

    def unregister_engine(self, engine_or_name):

        """Unregisters a mayavi engine specified either as a name or an
        engine instance."""

        engines = self.engines
        name = None
        if isinstance(engine_or_name, str):
            name = engine_or_name
        else:
            for key, engine in engines.items():
                if engine_or_name == engine:
                    name = key
                    break

        if name:
            del engines[name]
        logger.debug('Engine named %s unregistered', name)

    def get_file_reader(self, filename):

        """Given a filename, find a suitable source metadata that will
        read the file.

        Returns a suitable source metadata object that will
        handle this.
        """
        base, ext = splitext(filename)
        result = []
        if len(ext) > 0:
            ext = ext[1:]
            result = [src for src in self.sources \
                      if ext in src.extensions]

        # 'result' contains list of all source metadata that can handle
        # the file.

        # If there is only single source metadata available to handle
        # the file, we simply return it.

        # If there is a conflict i.e. more then one source metadata objects
        # capable of handling the file then we check if they are capable of
        # actually reading it using 'can_read_function' which may be a class
        # method or a simple function which returns whether the object is
        # capable of reading the file or not.

        # Finally returns the most suitable source metadata object to the engine. If
        # multiple objects are still present we return the last one in the list.

        if len(result) > 1:
            for res in result[:]:
                if len(res.can_read_test) > 0:
                    can_read = import_symbol(res.can_read_test)(filename)
                    if can_read:
                        return res
                    else:
                        result.remove(res)

        if len(result) == 0:
            return None

        return result[-1]

    def find_scene_engine(self, scene):
        """ Find the engine corresponding to a given tvtk scene.
        """
        for engine in self.engines.values():
            for s in engine.scenes:
                if scene is s:
                    return engine
                sc = s.scene
                if scene is sc:
                    return engine
                elif hasattr(sc, 'scene_editor') and \
                     scene is sc.scene_editor:
                    # This check is needed for scene model objects.
                    return engine
        else:
            raise TypeError("Scene not attached to a mayavi engine.")



# The global registry instance.
registry = Registry()

# Import the metadata from the sources, modules and filters so they are
# all registered.
from mayavi.sources.metadata import sources
registry.sources.extend(sources)

from mayavi.filters.metadata import filters
registry.filters.extend(filters)

from mayavi.modules.metadata import modules
registry.modules.extend(modules)

# Do any customizations from either the `site_mayavi.py` or the
# `user_mayavi.py` files.  This is done by importing the customize.py
# module here which in turn imports the necessary code from the users's
# customization.
from mayavi.core import customize



""" Defines the publicly accessible MayaVi2 sources.
"""

# Author: Frederic Petit, Prabhu Ramachandran
# Copyright (c) 2007-2020, Enthought, Inc.
# License: BSD Style.


from .array_source import ArraySource
from .builtin_image import BuiltinImage
from .builtin_surface import BuiltinSurface
from .chaco_reader import ChacoReader
from .image_reader import ImageReader
from .parametric_surface import ParametricSurface
from .plot3d_reader import PLOT3DReader
from .point_load import PointLoad
from .poly_data_reader import PolyDataReader
from .three_ds_importer import ThreeDSImporter
from .vrml_importer import VRMLImporter
from .volume_reader import VolumeReader
from .vtk_data_source import VTKDataSource
from .vtk_file_reader import VTKFileReader
from .vtk_xml_file_reader import VTKXMLFileReader
from .unstructured_grid_reader import UnstructuredGridReader



import unittest
from numpy import testing as npt
import numpy as np
from tvtk.api import tvtk

from mayavi import mlab

from mayavi.tests.common import get_example_data
from .test_mlab_integration import TestMlabNullEngine


def make_mb_dataset():
    fpath = get_example_data('pyramid_ug.vtu')
    r = tvtk.XMLUnstructuredGridReader(file_name=fpath)
    r.update()
    ug0 = r.output
    ug1 = tvtk.UnstructuredGrid()
    ug1.deep_copy(ug0)
    pts = ug1.points.to_array()
    pts[:, 0] += 10.0
    ug1.points = pts
    mb = tvtk.MultiBlockDataSet()
    mb.set_block(0, ug0)
    mb.set_block(1, ug1)
    return mb


class TestCompositeData(TestMlabNullEngine):
    def test_mb_contour(self):
        # Given
        mb = make_mb_dataset()

        # When
        src = mlab.pipeline.add_dataset(mb)
        iso = mlab.pipeline.iso_surface(src)

        # Then
        bounds = iso.actor.actor.bounds
        self.assertTrue((bounds[1] > 6) and (bounds[1] < 17))
        self.assertEqual(bounds[0], 0)
        self.assertEqual(bounds[2], 0)
        self.assertEqual(bounds[4], 0)
        self.assertTrue(bounds[5] < 3.1)

    def test_mb_surface(self):
        # Given
        mb = make_mb_dataset()

        # When
        src = mlab.pipeline.add_dataset(mb)
        surf = mlab.pipeline.surface(src)

        # Then
        bounds = surf.actor.actor.bounds
        expect = (0.0, 16.0, 0.0, 6.0, 0.0, 3.0)
        npt.assert_almost_equal(bounds, expect)

    def test_mb_scalar_cut_plane(self):
        # Given
        mb = make_mb_dataset()

        # When
        src = mlab.pipeline.add_dataset(mb)
        scp = mlab.pipeline.scalar_cut_plane(src)
        scp.implicit_plane.widget.normal_to_y_axis = True

        # Then
        bounds = scp.actor.actor.bounds
        expect = (0.0, 16.0, 3.0, 3.0, 0.0, 3.0)
        npt.assert_allclose(bounds, expect)

    def test_mb_vector_cut_plane(self):
        # Given
        mb = make_mb_dataset()

        # When
        src = mlab.pipeline.add_dataset(mb)
        vcp = mlab.pipeline.vector_cut_plane(src)
        vcp.implicit_plane.widget.normal_to_y_axis = True

        # Then
        bounds = vcp.actor.actor.bounds
        expect = np.array((0.0, 16.0, 3.0, 3.0, 0.0, 3.0))
        # print(expect - 1.0, expect + 1.0, bounds)
        self.assertTrue(
            np.all(bounds > (expect - 1.0)) and np.all(bounds < (expect + 1.0))
        )


if __name__ == '__main__':
    unittest.main()



"""Takes a cut plane of any input data set using an implicit plane and
plots the data with optional contouring and scalar warping.

"""
# Author: Prabhu Ramachandran <prabhu_r@users.sf.net>
# Copyright (c) 2005, Enthought, Inc.
# License: BSD Style.

# Enthought library imports.
from traits.api import Instance, Bool
from traitsui.api import View, Group, Item, InstanceEditor

# Local imports
from mayavi.core.module import Module
from mayavi.core.pipeline_info import PipelineInfo
from mayavi.components.implicit_plane import ImplicitPlane
from mayavi.components.cutter import Cutter
from mayavi.filters.warp_scalar import WarpScalar
from mayavi.components.poly_data_normals import PolyDataNormals
from mayavi.components.contour import Contour
from mayavi.components.actor import Actor


######################################################################
# `ScalarCutPlane` class.
######################################################################
class ScalarCutPlane(Module):

    # The version of this class.  Used for persistence.
    __version__ = 0

    # The implicit plane widget used to place the implicit function.
    implicit_plane = Instance(ImplicitPlane, allow_none=False,
                              record=True)

    # The cutter.  Takes a cut of the data on the implicit plane.
    cutter = Instance(Cutter, allow_none=False, record=True)

    # Specifies if contouring is to be done or not.
    enable_contours = Bool(False, desc='if contours are generated')

    # The Contour component that contours the data.
    contour = Instance(Contour, allow_none=False, record=True)

    # Specifies if scalar warping is to be done or not.
    enable_warp_scalar = Bool(False, desc='if scalar warping is enabled')

    # The WarpScalarCutPlane component that warps the data.
    warp_scalar = Instance(WarpScalar, allow_none=False, record=True)

    # Specify if scalar normals are to be computed to make a smoother surface.
    compute_normals = Bool(False, desc='if normals are to be computed '\
                           'to make the warped scalar surface smoother')

    # The component that computes the scalar normals.
    normals = Instance(PolyDataNormals, allow_none=False, record=True)

    # The actor component that represents the visualization.
    actor = Instance(Actor, allow_none=False, record=True)

    input_info = PipelineInfo(datasets=['any'],
                              attribute_types=['any'],
                              attributes=['scalars'])

    ########################################
    # View related code.

    _warp_group = Group(Item(name='filter',
                             style='custom',
                             editor=\
                             InstanceEditor(view=
                                            View(Item('scale_factor')))),
                        show_labels=False)

    view = View(Group(Item(name='implicit_plane',
                           style='custom'),
                      label='ImplicitPlane',
                      show_labels=False),
                Group(Group(Item(name='enable_contours')),
                      Group(Item(name='contour',
                                 style='custom',
                                 enabled_when='object.enable_contours'),
                            show_labels=False),
                      label='Contours',
                      show_labels=False),
                Group(Item(name='enable_warp_scalar'),
                      Group(Item(name='warp_scalar',
                                 enabled_when='enable_warp_scalar',
                                 style='custom',
                                 editor=InstanceEditor(view=
                                                       View(_warp_group))
                                 ),
                            show_labels=False,
                            ),
                      Item(name='_'),
                      Item(name='compute_normals',
                           enabled_when='enable_warp_scalar'),
                      Item(name='normals',
                           style='custom',
                           show_label=False,
                           enabled_when='compute_normals and enable_warp_scalar'),
                      label='WarpScalar',
                      show_labels=True),
                Group(Item(name='actor',
                           style='custom'),
                      label='Actor',
                      show_labels=False)
                )

    ######################################################################
    # `Module` interface
    ######################################################################
    def setup_pipeline(self):
        """Override this method so that it *creates* the tvtk
        pipeline.

        This method is invoked when the object is initialized via
        `__init__`.  Note that at the time this method is called, the
        tvtk data pipeline will *not* yet be setup.  So upstream data
        will not be available.  The idea is that you simply create the
        basic objects and setup those parts of the pipeline not
        dependent on upstream sources and filters.  You should also
        set the `actors` attribute up at this point.
        """
        # Create the objects.
        self.implicit_plane = ImplicitPlane()
        self.cutter = Cutter()
        self.contour = Contour(auto_contours=True, number_of_contours=10)
        self.warp_scalar = WarpScalar()
        self.normals = PolyDataNormals()
        self.actor = Actor()

        # Setup the actor suitably for this module.
        prop = self.actor.property
        prop.line_width = 2.0

    def update_pipeline(self):
        """Override this method so that it *updates* the tvtk pipeline
        when data upstream is known to have changed.

        This method is invoked (automatically) when any of the inputs
        sends a `pipeline_changed` event.
        """
        mm = self.module_manager
        if mm is None:
            return

        # Data is available, so set the input for the grid plane.
        self.implicit_plane.inputs = [mm.source]

        # Ensure that the warped scalar surface's normal is setup right.
        self.warp_scalar.filter.normal = self.implicit_plane.normal

        # This makes sure that any changes made to enable_warp when
        # the module is not running are updated when it is started --
        # this in turn calls the other functions (normals and
        # contours) internally.
        self._enable_warp_scalar_changed(self.enable_warp_scalar)

        # Set the LUT for the mapper.
        self.actor.set_lut(mm.scalar_lut_manager.lut)

        self.pipeline_changed = True

    def update_data(self):
        """Override this method so that it flushes the vtk pipeline if
        that is necessary.

        This method is invoked (automatically) when any of the inputs
        sends a `data_changed` event.
        """
        # Just set data_changed, the components should do the rest if
        # they are connected.
        self.data_changed = True

    ######################################################################
    # Non-public methods.
    ######################################################################
    def _get_warp_output(self):
        """Helper function to return the warped (or not) output
        depending on settings.
        """
        if self.enable_warp_scalar:
            if self.compute_normals:
                return self.normals
            else:
                return self.warp_scalar
        else:
            return self.cutter

    def _get_contour_output(self):
        """Helper function to return the contoured (and warped (or
        not)) output depending on settings.
        """
        if self.enable_contours:
            return self.contour
        else:
            return self._get_warp_output()

    def _filled_contours_changed_for_contour(self, value):
        """When filled contours are enabled, the mapper should use the
        the cell data, otherwise it should use the default scalar
        mode.
        """
        if value:
            self.actor.mapper.scalar_mode = 'use_cell_data'
        else:
            self.actor.mapper.scalar_mode = 'default'
        self.render()

    def _enable_warp_scalar_changed(self, value):
        """Turns on and off the scalar warping."""
        if self.module_manager is None:
            return

        if value:
            self.warp_scalar.inputs = [self.cutter]
        else:
            self.warp_scalar.inputs = []
        self._compute_normals_changed(self.compute_normals)
        self.render()

    def _compute_normals_changed(self, value):
        if self.module_manager is None:
            return

        if self.enable_warp_scalar:
            normals = self.normals
            if value:
                normals.inputs = [self.warp_scalar]
            else:
                normals.inputs = []
        self._enable_contours_changed(self.enable_contours)
        self.render()

    def _enable_contours_changed(self, value):
        """Turns on and off the contours."""
        if self.module_manager is None:
            return

        actor = self.actor
        if value:
            self.contour.inputs = [self._get_warp_output()]
            actor.inputs = [self._get_contour_output()]
            if self.contour.filled_contours:
                actor.mapper.scalar_mode = 'use_cell_data'
        else:
            self.contour.inputs = []
            actor.inputs = [self._get_warp_output()]
            actor.mapper.scalar_mode = 'default'
        self.render()

    def _normals_changed(self, old, new):
        warp_scalar = self.warp_scalar
        if warp_scalar is not None:
            new.inputs = [warp_scalar]
            self._compute_normals_changed(self.compute_normals)
        self._change_components(old, new)

    def _implicit_plane_changed(self, old, new):
        cutter = self.cutter
        if cutter is not None:
            cutter.cut_function = new.plane
            cutter.inputs = [new]
            # Update the pipeline.
            self._enable_warp_scalar_changed(self.enable_warp_scalar)
        # Hook up events to set the normals of the warp filter.
        if old is not None:
            old.widget.on_trait_change(self._update_normal, 'normal', remove=True)
        new.widget.on_trait_change(self._update_normal, 'normal')
        self._change_components(old, new)

    def _cutter_changed(self, old, new):
        ip = self.implicit_plane
        if ip is not None:
            new.cut_function = ip.plane
            new.inputs = [ip]
            # Update the pipeline.
            self._enable_warp_scalar_changed(self.enable_warp_scalar)
        self._change_components(old, new)

    def _contour_changed(self, old, new):
        # Update the pipeline.
        self._enable_contours_changed(self.enable_contours)
        self._change_components(old, new)

    def _warp_scalar_changed(self, old, new):
        # Update the pipeline.
        self._enable_warp_scalar_changed(self.enable_warp_scalar)
        self._change_components(old, new)

    def _actor_changed(self, old, new):
        # Update the pipeline.
        self._enable_contours_changed(self.enable_contours)
        self._change_components(old, new)

    def _update_normal(self):
        """Invoked when the orientation of the implicit plane changes.
        """
        ws = self.warp_scalar
        if ws is not None:
            ws.filter.normal = self.implicit_plane.widget.normal

